import numpy as np
import tlc
import torch
from ultralytics.models.yolo.segment.val import SegmentationValidator
from ultralytics.utils import ops

from tlc_ultralytics.constants import (
    IMAGE_COLUMN_NAME,
    SEGMENTATION_LABEL_COLUMN_NAME,
)
from tlc_ultralytics.detect.validator import TLCDetectionValidator
from tlc_ultralytics.utils.dataset import check_tlc_dataset


class TLCSegmentationValidator(TLCDetectionValidator, SegmentationValidator):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = SEGMENTATION_LABEL_COLUMN_NAME

    def check_dataset(self, *args, **kwargs):
        return check_tlc_dataset(*args, task="segment", settings=self._settings, **kwargs)

    def init_metrics(self, model):
        """Initialize metrics and use native mask processing for full-size masks."""
        super().init_metrics(model)
        # Always use native mask processing for 3LC to get full-size masks
        self.process = ops.process_mask_native

    def _get_metrics_schemas(self) -> dict[str, tlc.Schema]:
        # TODO: Ensure class  mapping is the same as in input table
        instance_properties_structure = {
            tlc.LABEL: tlc.CategoricalLabel(name=tlc.LABEL, classes=self.data["names_3lc"]),
            tlc.CONFIDENCE: tlc.Float(name=tlc.CONFIDENCE, number_role=tlc.NUMBER_ROLE_CONFIDENCE),
        }

        segment_sample_type = tlc.InstanceSegmentationMasks(
            name=tlc.PREDICTED_SEGMENTATIONS,
            instance_properties_structure=instance_properties_structure,
            is_prediction=True,
        )

        return {tlc.PREDICTED_SEGMENTATIONS: segment_sample_type.schema}

    def _compute_3lc_metrics(self, preds, batch) -> dict[str, list[dict[str, any]]]:
        """Compute 3LC metrics for instance segmentation.

        :param preds: Predictions returned by YOLO segmentation model.
        :param batch: Batch of data presented to the YOLO segmentation model.
        :returns: Metrics dict with predicted instance data for each sample in a batch.
        """
        predicted_batch_segmentations = []

        # Reimplements SegmentationValidator, but with control over mask processing
        for i, pred in enumerate(preds):
            pbatch = self._prepare_batch(i, batch)

            conf = pred["conf"]
            keep_indices = conf >= self._settings.conf_thres
            if not torch.any(keep_indices):
                height, width = pbatch["ori_shape"]
                predicted_instances = {
                    tlc.IMAGE_HEIGHT: height,
                    tlc.IMAGE_WIDTH: width,
                    tlc.INSTANCE_PROPERTIES: {
                        tlc.LABEL: [],
                        tlc.CONFIDENCE: [],
                    },
                    tlc.MASKS: np.zeros((height, width, 0), dtype=np.uint8),
                }
                predicted_batch_segmentations.append(predicted_instances)
                continue

            # Filter out low confidence predictions
            pred_conf = conf[keep_indices].tolist()
            pred_cls = pred["cls"][keep_indices].tolist()
            predicted_labels = [self.data["range_to_3lc_class"][int(p)] for p in pred_cls]
            predicted_masks = pred["masks"].clone()[keep_indices]

            # Get masks in resized dimensions (scale_masks expects (N, C, H, W) tensor)
            coco_masks = (
                ops.scale_masks(
                    predicted_masks[None],  # (N_masks, H, W) -> (1, N_masks, H, W)
                    pbatch["ori_shape"],
                    ratio_pad=pbatch["ratio_pad"],
                )[0]
                .byte()
                .cpu()
                .numpy()
            )
            coco_masks = np.transpose(coco_masks, (1, 2, 0))  # (N_masks, H, W) -> (H, W, N_masks)

            predicted_instances = {
                tlc.IMAGE_HEIGHT: pbatch["ori_shape"][0],
                tlc.IMAGE_WIDTH: pbatch["ori_shape"][1],
                tlc.INSTANCE_PROPERTIES: {
                    tlc.LABEL: predicted_labels,
                    tlc.CONFIDENCE: pred_conf,
                },
                tlc.MASKS: coco_masks,
            }

            predicted_batch_segmentations.append(predicted_instances)

        return {tlc.PREDICTED_SEGMENTATIONS: predicted_batch_segmentations}



from __future__ import annotations

from typing import Any

import cv2
import numpy as np
from tlc.core.builtins.constants import (
    CENTER_X,
    CENTER_Y,
    IMAGE,
    INSTANCES,
    INSTANCES_ADDITIONAL_DATA,
    LABEL,
    ORIENTED_BBS_2D,
    ROTATION,
    SIZE_X,
    SIZE_Y,
    X_MAX,
    X_MIN,
    Y_MAX,
    Y_MIN,
)

from tlc_ultralytics.detect.dataset import BaseTLCYOLODataset


def xcyxwhr_to_corner_points(xc, yc, w, h, r):
    # Ensure r is in the first quadrant [0, pi/2] by adding or subtracting pi/2 as needed
    if r < 0 or r > np.pi / 2:
        # Bring r into [0, pi/2] by adding or subtracting multiples of pi/2
        r = r % np.pi
        if r > np.pi / 2:
            r -= np.pi / 2
        elif r < 0:
            r += np.pi / 2
    assert 0 <= r <= np.pi / 2, f"Rotation r (in radians) must be in the first quadrant [0, pi/2]. Got {r}"
    dx, dy = w / 2.0, h / 2.0
    cos_r, sin_r = np.cos(r), np.sin(r)

    # local corners (relative to center): (±dx, ±dy)
    corners = np.array([[dx, dy], [dx, -dy], [-dx, -dy], [-dx, dy]], dtype=np.float64)

    R = np.array([[cos_r, -sin_r], [sin_r, cos_r]], dtype=np.float64)

    pts = corners @ R.T
    pts[:, 0] += xc
    pts[:, 1] += yc
    return pts


def corner_points_to_xywh(corner_points):
    pts = np.asarray(corner_points, dtype=np.float64).reshape(-1, 2)
    xmin, xmax = pts[:, 0].min(), pts[:, 0].max()
    ymin, ymax = pts[:, 1].min(), pts[:, 1].max()
    w, h = xmax - xmin, ymax - ymin
    xc, yc = xmin + w / 2.0, ymin + h / 2.0
    return np.array([xc, yc, w, h], dtype=np.float32)


def regularize_rboxes(rboxes):
    """
    Regularize rotated bounding boxes to range [0, pi/2].

    Args:
        rboxes (np.array): Input rotated boxes with shape (N, 5) in xywhr format.

    Returns:
        (np.array): Regularized rotated boxes.
    """
    x, y, w, h, t = rboxes.tolist()
    # Swap edge if t >= pi/2 while not being symmetrically opposite
    swap = t % np.pi >= np.pi / 2
    w_ = np.where(swap, h, w)
    h_ = np.where(swap, w, h)
    t = t % (np.pi / 2)
    return np.array([x, y, w_, h_, t], dtype=np.float32).reshape(1, 5)


class TLCOBBDataset(BaseTLCYOLODataset):
    """3LC YOLO dataset for OBB (oriented bounding boxes) models.

    Builds YOLO-compatible per-image labels dict with keys: im_file, shape, cls, bboxes.
    """

    def __init__(
        self,
        table,
        data=None,
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        super().__init__(
            table,
            data=data,
            exclude_zero=exclude_zero,
            class_map=class_map,
            image_column_name=image_column_name or IMAGE,
            label_column_name=label_column_name or ORIENTED_BBS_2D,
            **kwargs,
        )
        self._post_init()

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> dict[str, Any]:
        label_root = self._label_column_name.split(".")[0]
        label_column_value = row[label_root]

        x_min = label_column_value[X_MIN]
        y_min = label_column_value[Y_MIN]
        x_max = label_column_value[X_MAX]
        y_max = label_column_value[Y_MAX]
        image_width = x_max - x_min
        image_height = y_max - y_min

        # Get classes
        cls_arr = np.array(
            label_column_value[INSTANCES_ADDITIONAL_DATA][LABEL],
            dtype=np.float32,
        ).reshape(-1, 1)

        # Get bboxes
        boxes = []

        segments = []
        for instance in label_column_value[INSTANCES]:
            xc = instance[ORIENTED_BBS_2D][0][CENTER_X] / image_width
            yc = instance[ORIENTED_BBS_2D][0][CENTER_Y] / image_height
            w = instance[ORIENTED_BBS_2D][0][SIZE_X] / image_width
            h = instance[ORIENTED_BBS_2D][0][SIZE_Y] / image_height
            r = instance[ORIENTED_BBS_2D][0][ROTATION]

            r = r * 180 / np.pi  # Convert radians to degrees
            corner_points = cv2.boxPoints(((xc, yc), (w, h), r))
            segments.append(corner_points)
            box = corner_points_to_xywh(corner_points)
            boxes.append(box)

        bboxes_arr = np.array(boxes, ndmin=2, dtype=np.float32)
        if len(boxes) == 0:
            bboxes_arr = bboxes_arr.reshape(0, 4)

        return {
            "im_file": im_file,
            "shape": (round(image_height), round(image_width)),
            "cls": cls_arr,
            "bboxes": bboxes_arr,
            "keypoints": None,
            "segments": segments,
            "normalized": True,
            "bbox_format": "xywh",
            "example_id": example_id,
        }



from __future__ import annotations

import hashlib
import json
from collections.abc import Iterator
from multiprocessing.pool import ThreadPool
from typing import Any

import tlc
from ultralytics.data.utils import verify_image
from ultralytics.utils import LOGGER, NUM_THREADS, TQDM, colorstr


# Responsible for any generic 3LC dataset handling, such as scanning, caching and adding example ids to each sample
# Assume there is an attribute self.table that is a tlc.Table
class TLCDatasetMixin:
    def _post_init(self):
        self.display_name = self.table.dataset_name

        assert hasattr(self, "table") and isinstance(self.table, tlc.Table), (
            "TLCDatasetMixin requires an attribute `table` which is a tlc.Table."
        )

        if len(self.table) == 0:
            msg = f"The Table with URL {self.table.url.to_str()} has no rows, provide a Table populated with data."
            raise ValueError(msg)

    def __getitem__(self, index):
        """Get the item at the given index, add the example id to the sample for use in metrics collection."""
        example_id = self._index_to_example_id(index)
        sample = super().__getitem__(index)
        sample["example_id"] = example_id
        return sample

    @staticmethod
    def _absolutize_image_url(image_str: str, table_url: tlc.Url) -> str:
        """Expand aliases in the raw image string and absolutize the URL if it is relative.

        :param image_str: The raw image string to absolutize.
        :param table_url: The table URL to use for absolutization, usually the table whose images are being used.
        :return: The absolutized image string.
        :raises ValueError: If the alias cannot be expanded or the image URL is not a local file path.
        """
        url = tlc.Url(image_str)
        try:
            url = url.expand_aliases(allow_unexpanded=False)
        except ValueError as e:
            msg = f"Failed to expand alias in image_str: {image_str}. "
            msg += "Make sure the alias is spelled correctly and is registered in your configuration."
            raise ValueError(msg) from e

        if url.scheme not in (tlc.Scheme.FILE, tlc.Scheme.RELATIVE):
            msg = f"Image URL {url.to_str()} is not a local file path, it has scheme {url.scheme.value}. "
            msg += "Only local image file paths are supported. If your image URLs are not local, first copy "
            msg += "the images to a local directory and use an alias."
            raise ValueError(msg)

        return url.to_absolute(table_url).to_str()

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> Any:
        raise NotImplementedError("Subclasses must implement this method")

    def _index_to_example_id(self, index: int) -> int:
        raise NotImplementedError("Subclasses must implement this method")

    def _get_cache_key(self, image_paths: list[str]) -> str:
        """Generate a cache key based on the hash of all image paths.

        :param image_paths: List of absolute image paths
        :return: Cache key string
        """
        # Sort paths to ensure consistent hash regardless of order
        sorted_paths = sorted(image_paths)

        # Create hash of all paths concatenated
        paths_str = "".join(sorted_paths)
        return hashlib.md5(paths_str.encode()).hexdigest()

    def _get_cache_path(self, table_url: tlc.Url, cache_key: str) -> tlc.Url:
        """Get the URL to the cache file, in the same directory as the table.

        :param table_url: The table URL to use for the cache path
        :param cache_key: The cache key to use for the cache path
        :return: The URL to the cache file
        """
        return table_url / f"yolo_{cache_key}.json"

    def _load_cached_example_ids(self, cache_url: tlc.Url) -> list[int] | None:
        """Load the cached corrupt example ids from the cache file.

        :param cache_url: The path to the cache file
        :return: A list of corrupt example ids, or None if cache is invalid
        """
        try:
            cache_data = json.loads(cache_url.read(mode="s"))

            # Check cache version
            if cache_data.get("version") != 1:
                LOGGER.info("Cache version mismatch, regenerating cache.")
                return None

            if "corrupt_example_ids" not in cache_data:
                LOGGER.warning("Cache file missing corrupt_example_ids field, regenerating cache.")
                return None

            # Get corrupt example IDs
            corrupt_example_ids = cache_data["corrupt_example_ids"]
            return corrupt_example_ids

        except (json.JSONDecodeError, KeyError, ValueError) as e:
            LOGGER.warning(f"Failed to load cache: {e}, regenerating cache.")
            return None

    def _save_cached_example_ids(self, cache_url: tlc.Url, corrupt_example_ids: list[int]):
        """Save the corrupt example ids to the cache file.

        :param cache_url: The URL to the cache file
        :param corrupt_example_ids: A list of corrupt example ids
        """
        content = {
            "version": 1,
            "corrupt_example_ids": corrupt_example_ids,
        }

        cache_url.write(json.dumps(content, indent=2), mode="t")

    def _get_rows_from_table(self) -> tuple[list[str], list[Any]]:
        """Get the rows from the table and return a list of example ids, excluding zero weight and corrupt images.
        Rely on the cache to avoid recomputing example ids if possible.

        :return: A list of image paths and labels.
        """

        image_paths = [
            self._absolutize_image_url(row[self._image_column_name], self.table.url) for row in self.table.table_rows
        ]

        cache_key = self._get_cache_key(image_paths)
        cache_path = self._get_cache_path(self.table.url, cache_key)

        corrupt_example_ids = self._load_cached_example_ids(cache_path) if cache_path.exists() else None

        if corrupt_example_ids is not None:
            LOGGER.info(f"{colorstr(self.prefix)}: Loaded cached images.")

        if corrupt_example_ids is None:
            corrupt_example_ids = self._get_corrupt_example_ids_from_table(image_paths)
            self._save_cached_example_ids(cache_path, corrupt_example_ids)

        if len(corrupt_example_ids) == len(image_paths):
            msg = f"All images in the Table with URL {self.table.url.to_str()} are corrupt, can't use it."
            raise ValueError(msg)

        # Filter out corrupt and zero-weight example IDs
        example_ids = list(self._filter_example_ids(image_paths, corrupt_example_ids))

        if not example_ids:
            msg = (
                "No valid images found after filtering corrupt and zero-weight images in the Table with URL "
                f"{self.table.url.to_str()}. Please check the Table and ensure it contains valid images, or provide a "
                "Table with valid images."
            )
            raise ValueError(msg)

        im_files, labels = [], []
        for example_id in example_ids:
            im_file = image_paths[example_id]
            im_files.append(im_file)

            row = self.table.table_rows[example_id]
            labels.append(self._get_label_from_row(im_file, row, example_id))

        return im_files, labels

    def _filter_example_ids(self, image_paths: list[str], corrupt_example_ids: list[int]) -> Iterator[int]:
        """Filter example IDs to exclude corrupt and zero-weight images.

        :param image_paths: List of absolute image paths
        :param corrupt_example_ids: List of corrupt example IDs
        :yield: Valid example IDs
        """
        corrupt_set = set(corrupt_example_ids)
        weight_column_name = self.table.weights_column_name

        excluded_count = 0

        for example_id in range(len(image_paths)):
            # Skip corrupt images
            if example_id in corrupt_set:
                continue

            # Skip zero-weight images if exclusion is enabled
            if self._exclude_zero and self.table.table_rows[example_id].get(weight_column_name, 1) == 0:
                excluded_count += 1
                continue

            yield example_id

        if excluded_count > 0:
            percentage_excluded = excluded_count / len(self.table) * 100
            colored_prefix = colorstr(self.prefix + ":")
            LOGGER.info(
                f"{colored_prefix} Excluded {excluded_count} ({percentage_excluded:.2f}% of the table) "
                "zero-weight rows."
            )

    def _get_corrupt_example_ids_from_table(self, image_paths: list[str]) -> list[int]:
        """Get the corrupt example ids from the table by scanning all images.

        :param image_paths: List of absolute image paths
        :return: A list of corrupt example ids
        """
        corrupt_example_ids = []
        verified_count, corrupt_count, msgs = 0, 0, []
        colored_prefix = colorstr(self.prefix + ":")
        desc = f"{colored_prefix} Preparing data from {self.table.url.to_str()}"

        image_iterator = (((im_file, None), "") for im_file in image_paths)

        with ThreadPool(NUM_THREADS) as pool:
            results = pool.imap(func=verify_image, iterable=image_iterator)
            iterator = enumerate(results)
            pbar = TQDM(iterator, desc=desc, total=len(image_paths))

            for example_id, (_, verified, corrupt, msg) in pbar:
                if verified:
                    verified_count += 1
                elif corrupt:
                    corrupt_example_ids.append(example_id)
                    corrupt_count += 1

                if msg:
                    msgs.append(msg)

                pbar.desc = f"{desc} {verified_count} images, {corrupt_count} corrupt"

            pbar.close()

        if msgs:
            # Only take first 10 messages if there are more
            truncated = len(msgs) > 10
            msgs_to_show = msgs[:10]

            # Create the message string with truncation notice if needed
            msgs_str = "\n".join(msgs_to_show)
            if truncated:
                msgs_str += f"\n... (showing first 10 of {len(msgs)} messages)"

            percentage_corrupt = corrupt_count / len(image_paths) * 100

            verb = "is" if corrupt_count == 1 else "are"
            plural = "s" if corrupt_count != 1 else ""
            LOGGER.warning(
                f"{colored_prefix} There {verb} {corrupt_count} ({percentage_corrupt:.2f}%) corrupt image{plural}:"
                f"\n{msgs_str}"
            )

        return corrupt_example_ids



from __future__ import annotations

from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Literal

import tlc
import yaml
from tlc.core.builtins.constants import (
    INSTANCES_ADDITIONAL_DATA,
    LABEL,
)
from ultralytics.data.utils import check_cls_dataset, check_det_dataset
from ultralytics.utils import LOGGER, colorstr

from tlc_ultralytics.constants import TLC_COLORSTR, TLC_PREFIX
from tlc_ultralytics.settings import Settings


def get_dataset_functions(
    task: Literal["detect", "segment", "pose", "classify", "obb"],
) -> tuple[Callable, Callable, Callable]:
    if task == "detect":
        from tlc_ultralytics.detect.utils import (
            check_det_table,
            get_or_create_det_table,
        )

        dataset_checker = check_det_dataset
        table_creator = get_or_create_det_table
        table_checker = check_det_table
    elif task == "segment":
        from tlc_ultralytics.segment.utils import check_seg_table, get_or_create_seg_table

        dataset_checker = check_det_dataset
        table_creator = get_or_create_seg_table
        table_checker = check_seg_table
    elif task == "classify":
        from tlc_ultralytics.classify.utils import check_cls_table, get_or_create_cls_table

        dataset_checker = check_cls_dataset
        table_creator = get_or_create_cls_table
        table_checker = check_cls_table
    elif task == "pose":
        from tlc_ultralytics.pose.utils import check_pose_table, get_or_create_pose_table

        dataset_checker = check_det_dataset
        table_creator = get_or_create_pose_table
        table_checker = check_pose_table
    elif task == "obb":
        from tlc_ultralytics.obb.utils import check_obb_table, get_or_create_obb_table

        dataset_checker = check_det_dataset
        table_creator = get_or_create_obb_table
        table_checker = check_obb_table
    else:
        raise ValueError(f"Invalid task: {task}")
    return dataset_checker, table_creator, table_checker


def check_tlc_dataset(  # noqa: C901
    data: str,
    tables: dict[str, tlc.Table | tlc.Url | str] | None,
    image_column_name: str,
    label_column_name: str,
    project_name: str | None = None,
    splits: Iterable[str] | None = None,
    task: Literal["detect", "segment", "pose", "classify"] | None = None,
    settings: Settings | None = None,
) -> dict[str, tlc.Table | dict[float, str] | int]:
    """Get or create tables for YOLO datasets. data is ignored when tables is provided.

    :param data: Path to a dataset
    :param tables: Dictionary of tables, if already created
    :param image_column_name: Name of the column containing image paths
    :param label_column_name: Name of the column containing labels
    :param dataset_checker: Function to check the dataset (yolo implementation, download and checks)
    :param table_creator: Function to create the tables for the YOLO dataset
    :param table_checker: Function to check that a table is compatible with the current task
    :param project_name: Name of the project
    :param splits: List of splits to parse.
    :return: Dictionary of tables and class names
    """
    dataset_checker, table_creator, table_checker = get_dataset_functions(task)

    if not tables and not isinstance(data, (str, Path)):
        msg = "`data` must be a string. If you are passing tables directly, use the `tables` argument instead."
        raise ValueError(msg)

    if not tables and isinstance(data, str) and data.endswith(".ndjson"):
        msg = (
            "Using NDJson datasets directly is not supported in the YOLO integration. Create a tlc.Table from the "
            f'data with `tlc.Table.from_yolo_ndjson(ndjson_file="{data!s}", ...)` or convert it to a YOLO dataset and '
            "use `tlc.Table.from_yolo(...)`."
        )
        raise ValueError(msg)

    # If the data starts with the 3LC prefix, parse the YAML file and populate `tables`
    has_prefix = False
    if tables is None and isinstance(data, str) and data.startswith(TLC_PREFIX):
        has_prefix = True
        LOGGER.info(f"{TLC_COLORSTR}Parsing 3LC YAML file data={data} and populating tables")
        tables = parse_3lc_yaml_file(data)

    if tables is None:
        tables = {}

        data_dict = dataset_checker(data)

        # Get or create tables
        splits = splits or ("train", "val", "test", "minival")

        for key in splits:
            if data_dict.get(key):
                name = Path(data).stem
                dataset_name = f"{name}-{key}"
                table_name = "initial"

                if project_name is None:
                    project_name = f"{name}-YOLO"

                try:
                    table = table_creator(
                        key,
                        data_dict,
                        image_column_name=image_column_name,
                        label_column_name=label_column_name,
                        project_name=project_name,
                        dataset_name=dataset_name,
                        table_name=table_name,
                        settings=settings,
                    )

                    # Get the latest version when inferring
                    tables[key] = table.latest()

                    if tables[key] != table:
                        LOGGER.info(
                            f"{colorstr(key)}: Using latest version of table from {data}: "
                            f"{table.url} -> {tables[key].url}"
                        )
                    else:
                        LOGGER.info(f"{colorstr(key)}: Using initial version of table from {data}: {tables[key].url}")

                except Exception as e:
                    LOGGER.warning(
                        f"{colorstr(key)}: Failed to read or create table for split {key} from {data}: {e!s}"
                    )

    else:
        # LOGGER.info(f"{TLC_COLORSTR}Using data directly from tables")
        tables = tables.copy()
        _check_tables(tables)

        for key, table in tables.items():
            if splits is not None and key not in splits:
                continue

            if isinstance(table, (str, Path, tlc.Url)):
                try:
                    table_url = tlc.Url(table)
                    tables[key] = tlc.Table.from_url(table_url)
                except Exception as e:
                    raise ValueError(
                        f"Error loading table from {table} for split '{key}' provided through `tables`."
                    ) from e
            elif isinstance(table, tlc.Table):
                tables[key] = table
            else:
                msg = (
                    f"Invalid type {type(table)} for split {key} provided through `tables`."
                    "Must be a tlc.Table object or a location (string, pathlib.Path or tlc.Url) of a tlc.Table."
                )

                raise ValueError(msg)

            # Check that the table is compatible with the current task
            if table_checker is not None:
                table_checker(tables[key], image_column_name, label_column_name)

            source = "3LC YAML file" if has_prefix else "provided tables"
            LOGGER.info(f"{colorstr(key)}: Using table {tables[key].url} from {source}")

    first_split = next(iter(tables.keys()))

    value_map = get_value_map_from_table(tables[first_split], label_column_name, task)
    names = tlc.SchemaHelper.to_simple_value_map(value_map)
    if task == "pose":
        kpt_shape = tlc.KeypointHelper.get_keypoint_shape_from_table(tables[first_split], label_column_name)
        flip_idx = tlc.KeypointHelper.get_flip_indices_from_table(tables[first_split], label_column_name)
        keypoint_attributes = tlc.KeypointHelper.get_keypoint_attributes_from_table(
            tables[first_split], label_column_name
        )
        lines = tlc.KeypointHelper.get_lines_from_table(tables[first_split], label_column_name)
        line_attributes = tlc.KeypointHelper.get_line_attributes_from_table(tables[first_split], label_column_name)
        triangles = tlc.KeypointHelper.get_triangles_from_table(tables[first_split], label_column_name)
        triangle_attributes = tlc.KeypointHelper.get_triangle_attributes_from_table(
            tables[first_split], label_column_name
        )
        oks_sigmas = tlc.KeypointHelper.get_oks_sigmas_from_table(tables[first_split], label_column_name)
        points = tlc.KeypointHelper.get_points_from_table(tables[first_split], label_column_name)
    else:
        kpt_shape = [17, 3]  # yolo default
        flip_idx, keypoint_attributes, lines, line_attributes, triangles, triangle_attributes, points = (
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )

    if names is None:
        raise ValueError(f"Failed to get value map for table with Url: {tables[first_split].url}")

    for split, split_table in tables.items():
        if split == first_split:
            continue

        split_value_map = get_value_map_from_table(split_table, label_column_name, task)
        split_names = tlc.SchemaHelper.to_simple_value_map(split_value_map)

        if split_names is None:
            raise ValueError(f"Failed to get value map for table with Url: {tables[split].url}")

        if split_names != names:
            first_items = set(names.items())
            split_items = set(split_names.items())

            only_in_first = first_items - split_items
            only_in_split = split_items - first_items

            messages = []

            if only_in_first:
                dict_str = "{" + ", ".join(f"{k}: '{v}'" for k, v in only_in_first) + "}"
                messages.append(f"'{first_split}' has categories that '{split}' does not: {dict_str}")
            if only_in_split:
                dict_str = "{" + ", ".join(f"{k}: '{v}'" for k, v in only_in_split) + "}"
                messages.append(f"'{split}' has categories that '{first_split}' does not: {dict_str}")

            error_msg = "All splits must have the same categories, but " + " and ".join(messages)

            raise ValueError(error_msg)

    # Map name indices to 0, 1, ..., n-1
    names_yolo = dict(enumerate(names.values()))
    range_to_3lc_class = dict(enumerate(names))

    ret = {
        **tables,
        "names": names_yolo,
        "names_3lc": value_map,
        "nc": len(names),
        "range_to_3lc_class": range_to_3lc_class,
        "3lc_class_to_range": {v: k for k, v in range_to_3lc_class.items()},
        "channels": 3,  # TODO(Frederik): Read out channels from appropriate place and populate here
        "kpt_shape": kpt_shape,
    }
    if task == "pose":
        if flip_idx is not None:
            ret["flip_idx"] = flip_idx
        if keypoint_attributes is not None:
            ret["keypoint_attributes"] = keypoint_attributes
        if lines is not None:
            ret["lines"] = lines
        if line_attributes is not None:
            ret["line_attributes"] = line_attributes
        if triangles is not None:
            ret["triangles"] = triangles
        if triangle_attributes is not None:
            ret["triangle_attributes"] = triangle_attributes
        if oks_sigmas is not None:
            ret["oks_sigmas"] = oks_sigmas
        if points is not None:
            ret["points"] = points
    return ret


def get_value_map_from_table(
    table: tlc.Table,
    label_column_name: str,
    task: Literal["detect", "segment", "pose", "classify", "obb"],
) -> dict[int, str]:
    if task == "pose":
        try:
            return table.rows_schema[label_column_name][INSTANCES_ADDITIONAL_DATA][LABEL].value.map
        except Exception as e:
            raise ValueError("Failed to get value map from table") from e
    elif task == "obb":
        try:
            return table.rows_schema[label_column_name][INSTANCES_ADDITIONAL_DATA][LABEL].value.map
        except Exception as e:
            raise ValueError("Failed to get value map from table") from e
    else:
        return table.get_value_map(label_column_name)


def parse_3lc_yaml_file(data_file: str) -> dict[str, tlc.Table]:
    """Parse a 3LC YAML file and return the corresponding tables.

    :param data_file: The path to the 3LC YAML file.
    :returns: The tables pointed to by the YAML file.
    """
    # Read the YAML file, removing the prefix
    if not (data_file_url := tlc.Url(data_file.replace(TLC_PREFIX, ""))).exists():
        raise FileNotFoundError(f"Could not find YAML file {data_file_url}")

    data_config = yaml.safe_load(data_file_url.read())

    path = data_config.get("path")
    splits = [key for key in data_config if key != "path"]

    tables = {}
    for split in splits:
        # Handle :latest at the end
        if data_config[split].endswith(":latest"):
            latest = True
            split_path = data_config[split][: -len(":latest")]
        else:
            latest = False
            split_path = data_config[split]

        if split_path.startswith("./"):
            LOGGER.debug(f"{TLC_COLORSTR}{split} split path starts with './', removing it.")
            split_path = split_path[2:]

        table_url = tlc.Url(path) / split_path if path else tlc.Url(split_path)

        table = tlc.Table.from_url(table_url)

        if latest:
            table = table.latest()

        tables[split] = table

    return tables


def _check_tables(tables: object):
    if not isinstance(tables, dict):
        msg = f"When providing tables directly, they must be a dictionary, but got type {type(tables)}."
        raise ValueError(msg)

    for key, table in tables.items():
        if not isinstance(table, (str, Path, tlc.Url, tlc.Table)):
            msg = (
                "When providing tables directly, they must be a tlc.Table or a URL to a tlc.Table. ",
                f"Got {type(table)} for split {key}.",
            )
            raise ValueError(msg)



from __future__ import annotations

import tlc
from tlc.core.builtins.constants.column_names import INSTANCES, INSTANCES_ADDITIONAL_DATA, LABEL, ORIENTED_BBS_2D

from tlc_ultralytics.settings import Settings


def get_or_create_obb_table(
    key: str,
    data_dict: dict[str, object],
    image_column_name: str,
    label_column_name: str,
    project_name: str,
    dataset_name: str,
    table_name: str,
    settings: Settings | None = None,
) -> tlc.Table:
    return tlc.Table.from_yolo(
        dataset_yaml_file=data_dict["yaml_file"],
        split=key,
        override_split_path=data_dict[key],
        task="obb",
        project_name=project_name,
        dataset_name=dataset_name,
        table_name=table_name,
        if_exists="reuse",
        add_weight_column=True,
        description="Created with 3LC YOLO integration",
    )


def check_obb_table(table: tlc.Table, image_column_name: str, label_column_name: str) -> None:
    """Verify that the table is compatible with instance segmentation.

    :param table: The table to check.
    :param image_column_name: The name of the image column.
    :param label_column_name: The value path of the label.
    :raises ValueError: If the table is not compatible with instance segmentation.
    """
    row_schema = table.row_schema
    label_column_name = label_column_name.split(".")[0]
    try:
        assert image_column_name in row_schema, f"Image column '{image_column_name}' not found."
        assert label_column_name in row_schema, f"Label column '{label_column_name}' not found."

        assert INSTANCES in row_schema[label_column_name], f"Label column '{label_column_name}' missing instances."
        assert ORIENTED_BBS_2D in row_schema[label_column_name][INSTANCES], (
            f"Label column '{label_column_name}' missing oriented bbs 2d."
        )
        assert INSTANCES_ADDITIONAL_DATA in row_schema[label_column_name], (
            f"Label column '{label_column_name}' missing instances additional data."
        )
        assert LABEL in row_schema[label_column_name][INSTANCES_ADDITIONAL_DATA], (
            f"Label column '{label_column_name}' missing label."
        )
    except (AssertionError, ValueError) as e:
        msg = f"Data validation failed for {label_column_name} column in table with URL {table.url}. {e!s}"
        raise ValueError(msg) from e



from __future__ import annotations

import tlc
from tlc.core.builtins.constants import (
    BBS_2D,
    INSTANCES,
    INSTANCES_ADDITIONAL_DATA,
    VERTICES_2D,
    X_MAX,
    X_MIN,
    Y_MAX,
    Y_MIN,
)

from tlc_ultralytics.settings import Settings


def get_or_create_pose_table(
    key: str,
    data_dict: dict[str, object],
    image_column_name: str,
    label_column_name: str,
    project_name: str,
    dataset_name: str,
    table_name: str,
    settings: Settings | None = None,
) -> tlc.Table:
    return tlc.Table.from_yolo(
        dataset_yaml_file=data_dict["yaml_file"],
        split=key,
        override_split_path=data_dict[key],
        task="pose",
        project_name=project_name,
        dataset_name=dataset_name,
        table_name=table_name,
        if_exists="reuse",
        add_weight_column=True,
        description="Created with 3LC YOLO integration",
        points=settings.points,
        lines=settings.lines,
        triangles=settings.triangles,
        point_attributes=settings.point_attributes,
        line_attributes=settings.line_attributes,
        triangle_attributes=settings.triangle_attributes,
        flip_indices=settings.flip_indices,
    )


def check_pose_table(table: tlc.Table, image_column_name: str, label_column_name: str) -> None:
    """Verify that the table is compatible with pose keypoints.

    :param table: The table to check.
    :param image_column_name: The name of the image column.
    :param label_column_name: The name of the pose label root column (e.g., 'pose').
    :raises ValueError: If the table is not compatible with pose.
    """
    row_schema = table.row_schema.values

    label_root = label_column_name.split(".")[0]

    try:
        assert image_column_name in row_schema, f"Image column '{image_column_name}' not found."
        assert label_root in row_schema, f"Pose column '{label_root}' not found."

        schema = row_schema[label_root]
        assert hasattr(schema, "values"), f"Pose column '{label_root}' has no values schema."
        for key in (X_MIN, Y_MIN, X_MAX, Y_MAX, INSTANCES, INSTANCES_ADDITIONAL_DATA):
            assert key in schema.values, f"Pose column '{label_root}' missing key '{key}'."

        instances_schema = schema.values[INSTANCES]
        assert hasattr(instances_schema, "values"), "Instances schema must be composite."
        for k in (VERTICES_2D, BBS_2D):
            assert k in instances_schema.values, f"Instances missing '{k}'."

    except (AssertionError, KeyError) as e:
        raise ValueError(f"Table with url {table.url} is not compatible with YOLO pose. {e}") from None


def yolo_pose_loss_schemas(training: bool = False) -> dict[str, tlc.Schema]:
    """Create 3LC schemas for YOLO pose per-sample loss metrics.

    :param training: Whether metrics are collected during training.
    :returns: The YOLO pose loss schemas for each component.
    """
    schemas: dict[str, tlc.Schema] = {}
    schemas["box_loss"] = tlc.Schema(
        description="Box Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3004,
    )
    schemas["dfl_loss"] = tlc.Schema(
        description="Distribution Focal Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3005,
    )
    schemas["cls_loss"] = tlc.Schema(
        description="Classification Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3006,
    )
    schemas["pose_loss"] = tlc.Schema(
        description="Keypoint location loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3008,
    )
    schemas["kobj_loss"] = tlc.Schema(
        description="Keypoint visibility/objectness loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3009,
    )
    if training:
        schemas["loss"] = tlc.Schema(
            description=(
                "Weighted sum of box, DFL, classification, keypoint location and visibility losses used in training"
            ),
            writable=False,
            value=tlc.Float32Value(),
            display_importance=3010,
        )
    return schemas



from __future__ import annotations

from typing import Any

import numpy as np
from tlc.core.builtins.constants import IMAGE, KEYPOINTS_2D
from tlc.core.data_formats import Keypoints2DInstances

from tlc_ultralytics.detect.dataset import BaseTLCYOLODataset


class TLCYOLOPoseDataset(BaseTLCYOLODataset):
    """3LC YOLO dataset for pose (keypoints) models.

    Builds YOLO-compatible per-image labels dict with keys: im_file, shape, cls, bboxes, keypoints.
    """

    def __init__(
        self,
        table,
        data=None,
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        super().__init__(
            table,
            data=data,
            exclude_zero=exclude_zero,
            class_map=class_map,
            image_column_name=image_column_name or IMAGE,
            label_column_name=label_column_name or KEYPOINTS_2D,
            **kwargs,
        )
        self._post_init()

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> dict[str, Any]:
        pose_root = self._label_column_name.split(".")[0]
        label_column_value = row[pose_root]

        # Desired fixed K from dataset config (default 17) for empty-case shapes
        kpt_shape = self.data.get("kpt_shape")
        instances = Keypoints2DInstances.from_row(label_column_value)

        # Image dimensions and raw arrays
        H = float(instances.image_height)
        W = float(instances.image_width)
        labels = instances.instance_labels.astype(np.int32, copy=False)  # (N,)
        bboxes_xyxy = instances.instance_bbs.astype(np.float32, copy=False)  # (N,4) [x_min, y_min, x_max, y_max]
        kxy = instances.keypoints.astype(np.float32, copy=False)  # (N,K,2)
        vis = instances.keypoint_visibilities  # (N,K) or None

        # Normalize keypoints to [0,1]
        kxy[..., 0] /= W
        kxy[..., 1] /= H

        # Vectorized: xyxy (abs) -> center-xywh (normalized)
        xy_min = bboxes_xyxy[:, 0:2]
        xy_max = bboxes_xyxy[:, 2:4]
        wh_abs = xy_max - xy_min
        ctr_abs = xy_min + 0.5 * wh_abs
        scale = np.array([W, H], dtype=np.float32)
        bboxes_arr = np.concatenate([(ctr_abs / scale), (wh_abs / scale)], axis=1).astype(np.float32)

        # Handle empty vs non-empty
        if labels.size == 0:
            K_cfg = int(kpt_shape[0]) if kpt_shape else (int(kxy.shape[1]) if kxy.shape[1] > 0 else 17)
            cls_arr = np.zeros((0, 1), dtype=np.float32)
            bboxes_arr = np.zeros((0, 4), dtype=np.float32)
            kp_stack = np.zeros((0, K_cfg, 3), dtype=np.float32)
        else:
            mapped_list = [self._class_map.get(int(v), int(v)) for v in labels.tolist()]
            cls_arr = np.asarray(mapped_list, dtype=np.float32).reshape(-1, 1)

            # Build (N,K,3) with visibilities
            if vis is None:
                vis_arr = np.ones((kxy.shape[0], kxy.shape[1], 1), dtype=np.float32)
            else:
                vis_arr = vis.astype(np.float32, copy=False).reshape(kxy.shape[0], kxy.shape[1], 1)
            kp_stack = np.concatenate([kxy, vis_arr], axis=2)

        return {
            "im_file": im_file,
            "shape": (round(instances.image_height), round(instances.image_width)),
            "cls": cls_arr,
            "bboxes": bboxes_arr,
            "segments": [],
            "keypoints": kp_stack,  # (N, K, 3)
            "normalized": True,
            "bbox_format": "xywh",
            "example_id": example_id,
        }



import numpy as np
import tlc
import torch
from tlc.core.builtins.constants import (
    CONFIDENCE,
    INSTANCES,
    INSTANCES_ADDITIONAL_DATA,
    LABEL,
    X_MAX,
    X_MIN,
    Y_MAX,
    Y_MIN,
)
from tlc.core.builtins.schemas import CategoricalLabelListSchema, Float32ListSchema, Geometry2DSchema
from ultralytics.models.yolo.obb.val import OBBValidator
from ultralytics.utils import ops

from tlc_ultralytics.constants import (
    IMAGE_COLUMN_NAME,
    OBB_LABEL_COLUMN_NAME,
)
from tlc_ultralytics.detect.validator import TLCDetectionValidator
from tlc_ultralytics.utils.dataset import check_tlc_dataset


class TLCOBBValidator(TLCDetectionValidator, OBBValidator):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = OBB_LABEL_COLUMN_NAME

    def check_dataset(self, *args, **kwargs):
        return check_tlc_dataset(*args, task="obb", settings=self._settings, **kwargs)

    def _get_metrics_schemas(self) -> dict[str, tlc.Schema]:
        return {
            "oriented_bbs_2d_predicted": Geometry2DSchema(
                include_2d_oriented_bounding_boxes=True,
                per_instance_schemas={
                    LABEL: CategoricalLabelListSchema(classes=self.data["names"]),
                    CONFIDENCE: Float32ListSchema(),
                },
            )
        }

    def _compute_3lc_metrics(self, preds, batch) -> dict[str, list[dict[str, any]]]:
        """Compute 3LC metrics for instance segmentation.

        :param preds: Predictions returned by YOLO segmentation model.
        :param batch: Batch of data presented to the YOLO segmentation model.
        :returns: Metrics dict with predicted instance data for each sample in a batch.
        """
        predicted = []

        for i, pred in enumerate(preds):
            predicted_bboxes, predicted_classes, predicted_confidences = (
                pred["bboxes"].clone(),
                pred["cls"].clone(),
                pred["conf"].clone(),
            )
            h, w = batch["ori_shape"][i]
            mask = predicted_confidences >= self._settings.conf_thres

            if len(pred) == 0 or not torch.any(mask):
                predicted.append(
                    {
                        X_MIN: 0,
                        Y_MIN: 0,
                        X_MAX: w,
                        Y_MAX: h,
                        INSTANCES: [],
                        INSTANCES_ADDITIONAL_DATA: {LABEL: [], CONFIDENCE: []},
                    }
                )
                continue

            predicted_bboxes = predicted_bboxes[mask]
            predicted_classes = predicted_classes[mask].cpu().numpy().astype(np.int32).tolist()
            predicted_confidences = predicted_confidences[mask].cpu().numpy().astype(np.float32).tolist()

            resized_shape = batch["resized_shape"][i]
            ori_shape = batch["ori_shape"][i]
            ratio_pad = batch["ratio_pad"][i]
            scaled_bboxes = ops.scale_boxes(resized_shape, predicted_bboxes, ori_shape, ratio_pad, xywh=True)

            instances = []
            for j in range(len(predicted_bboxes)):
                predicted_bbox = scaled_bboxes[j].cpu().numpy().astype(np.float32).tolist()
                instances.append(
                    {
                        "oriented_bbs_2d": [
                            {
                                "center_x": predicted_bbox[0],
                                "center_y": predicted_bbox[1],
                                "size_x": predicted_bbox[2],
                                "size_y": predicted_bbox[3],
                                "rotation": predicted_bbox[4],
                            }
                        ],
                    }
                )

            predicted.append(
                {
                    X_MIN: 0,
                    Y_MIN: 0,
                    X_MAX: w,
                    Y_MAX: h,
                    INSTANCES: instances,
                    INSTANCES_ADDITIONAL_DATA: {LABEL: predicted_classes, CONFIDENCE: predicted_confidences},
                }
            )

        return {"oriented_bbs_2d_predicted": predicted}



from __future__ import annotations

from collections.abc import Iterable

import tlc
import ultralytics
from ultralytics.models import yolo
from ultralytics.models.yolo.model import YOLO as YOLOBase
from ultralytics.nn.tasks import ClassificationModel, DetectionModel, OBBModel, PoseModel, SegmentationModel
from ultralytics.utils import LOGGER

from tlc_ultralytics.classify import (
    TLCClassificationTrainer,
    TLCClassificationValidator,
)
from tlc_ultralytics.constants import DEFAULT_COLLECT_RUN_DESCRIPTION
from tlc_ultralytics.detect import TLCDetectionTrainer, TLCDetectionValidator
from tlc_ultralytics.obb import TLCOBBTrainer, TLCOBBValidator
from tlc_ultralytics.pose import TLCPoseTrainer, TLCPoseValidator
from tlc_ultralytics.segment import TLCSegmentationTrainer, TLCSegmentationValidator
from tlc_ultralytics.settings import Settings
from tlc_ultralytics.utils import check_requirements, reduce_embeddings


class YOLO(YOLOBase):
    """YOLO (You Only Look Once) object detection model with 3LC integration."""

    def __init__(self, *args, **kwargs):
        """Initialize YOLO model with 3LC integration. Checks that the installed version of 3LC is compatible."""

        check_requirements()

        super().__init__(*args, **kwargs)

    def train(self, *args, **kwargs):
        """Train the model."""

        # Patch the check_pip_update_available function to avoid prompting for an update
        def check_pip_update_available_return_false():
            return False

        ultralytics_check_pip_update_available = ultralytics.utils.checks.check_pip_update_available
        ultralytics.utils.checks.check_pip_update_available = check_pip_update_available_return_false

        # Ensure 'model' key exists in overrides (may be cleared after previous train() call)
        if "model" not in self.overrides:
            self.overrides["model"] = self.model_name

        output = super().train(*args, **kwargs)

        # Restore the original function
        ultralytics.utils.checks.check_pip_update_available = ultralytics_check_pip_update_available

        return output

    @property
    def task_map(self):
        """Map head to 3LC model, trainer, validator, and predictor classes."""
        return {
            "detect": {
                "model": DetectionModel,
                "trainer": TLCDetectionTrainer,
                "validator": TLCDetectionValidator,
                "predictor": yolo.detect.DetectionPredictor,
            },
            "classify": {
                "model": ClassificationModel,
                "trainer": TLCClassificationTrainer,
                "validator": TLCClassificationValidator,
                "predictor": yolo.classify.ClassificationPredictor,
            },
            "segment": {
                "model": SegmentationModel,
                "trainer": TLCSegmentationTrainer,
                "validator": TLCSegmentationValidator,
                "predictor": yolo.segment.SegmentationPredictor,
            },
            "pose": {
                "model": PoseModel,
                "trainer": TLCPoseTrainer,
                "validator": TLCPoseValidator,
                "predictor": yolo.pose.PosePredictor,
            },
            "obb": {
                "model": OBBModel,
                "trainer": TLCOBBTrainer,
                "validator": TLCOBBValidator,
                "predictor": yolo.obb.OBBPredictor,
            },
        }

    def collect(
        self,
        data: str | None = None,
        splits: Iterable[str] | None = None,
        tables: dict[str, str | tlc.Url | tlc.Table] | None = None,
        settings: Settings | None = None,
        **kwargs,
    ) -> dict[str, dict[str, float]]:
        """Perform calls to model.val() to collect metrics on a set of splits, all under one tlc.Run.
        If enabled, embeddings are reduced at the end of validation.

        :param data: Path to a YOLO or 3LC YAML file. If provided, splits must also be provided.
        :param splits: List of splits to collect metrics for. If provided, data must also be provided.
        :param tables: Dictionary of splits to tables to collect metrics for. Mutually exclusive with data and splits.
        :param settings: 3LC settings to use for collecting metrics. If None, default settings are used.
        :param kwargs: Additional keyword arguments are forwarded as model.val(**kwargs).
        :return: Dictionary of split names to results returned by model.val().
        """
        # Verify only data+splits or tables are provided
        if not ((data and splits) or tables):
            raise ValueError("Either data and splits or tables must be provided to collect.")

        if settings is None:
            settings = Settings()

        if not settings.run_description:
            settings.run_description = DEFAULT_COLLECT_RUN_DESCRIPTION

        results_dict = {}
        # Call val for each split or table
        if data and splits:
            for split in splits:
                results_dict[split] = self.val(data=data, split=split, settings=settings, **kwargs)
        elif tables:
            for split in tables:
                results_dict[split] = self.val(table=tables[split], settings=settings, **kwargs)

        # Reduce embeddings
        if settings and settings.image_embeddings_dim > 0:
            # TODO: Allow user to pass in preferred foreign_table_url

            reduce_embeddings(
                tlc.active_run(),
                method=settings.image_embeddings_reducer,
                n_components=settings.image_embeddings_dim,
                reducer_args=settings.image_embeddings_reducer_args,
            )

        tlc.active_run().set_status_completed()

        return results_dict


class TLCYOLO(YOLO):
    def __init__(self, *args, **kwargs):
        LOGGER.warning("TLCYOLO is deprecated and will be removed in a future version. Use YOLO instead.")

        super().__init__(*args, **kwargs)



from __future__ import annotations

import json
from pathlib import Path

import tlc
import ultralytics
from ultralytics.engine.trainer import BaseTrainer
from ultralytics.utils import DEFAULT_CFG, LOGGER, RANK
from ultralytics.utils.metrics import smooth

from tlc_ultralytics.constants import (
    DEFAULT_TRAIN_RUN_DESCRIPTION,
    TLC_COLORSTR,
)
from tlc_ultralytics.engine.utils import (
    _complete_label_column_name,
    _handle_deprecated_column_name,
    _restore_random_state,
)
from tlc_ultralytics.settings import Settings
from tlc_ultralytics.utils import reduce_embeddings
from tlc_ultralytics.utils.generate_ddp import generate_ddp_command


class TLCTrainerMixin(BaseTrainer):
    """A class extending the BaseTrainer class for training Ultralytics YOLO models with 3LC,
    which implements common 3LC-specific behavior across tasks. Use as a Mixin class for task-specific
    trainers.
    """

    def __init__(self, cfg=DEFAULT_CFG, overrides=None, _callbacks=None):
        LOGGER.info("Using 3LC Trainer 🌟")

        if RANK == -1:
            # Settings
            self._settings = overrides.pop("settings", Settings())

            self._settings.image_column_name = _handle_deprecated_column_name(
                overrides.pop("image_column_name", None),
                self._settings.image_column_name,
                self._default_image_column_name,
                column_name="image_column_name",
            )
            self._settings.label_column_name = _handle_deprecated_column_name(
                overrides.pop("label_column_name", None),
                self._settings.label_column_name,
                self._default_label_column_name,
                column_name="label_column_name",
            )

            self._settings.label_column_name = _complete_label_column_name(
                self._settings.label_column_name,
                self._default_label_column_name,
            )

            self._settings.verify(training=True)

            # Tables / data
            if "data" not in overrides and "tables" not in overrides:
                msg = "You must provide either `data` or `tables` to train with 3LC."
                raise ValueError(msg)

            if "data" in overrides and not isinstance(overrides["data"], (str, Path)):
                msg = "`data` must be a string. If you are passing tables directly, use the `tables` argument instead."
                raise ValueError(msg)

            # Ensure tables is a dictionary of URLs to the table urls
            tables = overrides.pop("tables", None)
            if tables:
                self._tables = self._handle_tables_argument(tables)
            else:
                self._tables = None

        else:  # DDP worker nodes
            try:
                data = json.loads(overrides["data"])
            except json.JSONDecodeError as e:
                msg = f"RANK {RANK} expected `data` to be a JSON string with run url, settings and tables."
                raise ValueError(msg) from e

            for key in ("run_url", "settings"):
                if key not in data:
                    msg = f"RANK {RANK} expected `data` to contain a `{key}` key."
                    raise ValueError(msg)

            self._run = tlc.Run.from_url(data["run_url"])
            self._settings = Settings(**data["settings"])
            self._tables = data.get("tables", None)
            overrides["data"] = data.get("data", "")

        super().__init__(cfg, overrides, _callbacks)

        self._metrics_collection_epochs = set(self._settings.get_metrics_collection_epochs(self.epochs))
        self._train_validator = None
        self._train_equals_val_result = None

        if RANK == -1:
            self._create_run()
            self._log_3lc_parameters()
            self._run.set_status_running()
            self._print_metrics_collection_epochs()
            self._print_task_specific_parameters()

    def train(self):
        """Override the train method to use custom generate_ddp_command function to serialize 3LC data in data
        argument.
        """
        ultralytics.engine.trainer.generate_ddp_command = generate_ddp_command
        super().train()
        ultralytics.engine.trainer.generate_ddp_command = ultralytics.utils.dist.generate_ddp_command

    def _serialize_state(self) -> str:
        """Serialize the run url, settings and tables to a JSON string.

        :returns: A JSON string with the run url, settings and table URLs.
        """
        return json.dumps(
            {
                "run_url": self._run.url.to_str(),
                "settings": self._settings.to_dict(),
                "data": self.args.data,
                "tables": self._tables if self._tables else None,
            }
        )

    def _create_run(self) -> None:
        """Create a run."""
        # Create a 3LC run
        description = (
            self._settings.run_description if self._settings.run_description else DEFAULT_TRAIN_RUN_DESCRIPTION
        )

        project_name = self._settings.project_name if self._settings.project_name else self.data["train"].project_name
        self._run = tlc.init(
            project_name=project_name,
            description=description,
            run_name=self._settings.run_name,
        )

        LOGGER.info(f"{TLC_COLORSTR}Created run named '{self._run.url.parts[-1]}' in project {self._run.project_name}.")

    @staticmethod
    def _handle_tables_argument(tables: dict[str, tlc.Table | str | Path | tlc.Url]) -> dict[str, str]:
        """Handle the tables argument and return it as a dictionary of URLs (as strings) to the Tables."""
        table_urls = {}

        for k, v in tables.items():
            if isinstance(v, tlc.Table):
                table_urls[k] = v.url.to_str()
            elif isinstance(v, (str, Path, tlc.Url)):
                table_urls[k] = tlc.Url(v).to_str()
            else:
                raise ValueError(f"Invalid type {type(v)} for split {k} of tables provided directly.")

        return table_urls

    def _log_3lc_parameters(self):
        """Log various data as parameters to the tlc.Run."""
        if "val" in self.data:
            val_url = str(self.data["val"].url)
        else:
            val_url = str(self.data["test"].url)

        parameters = {
            **vars(self.args),  # YOLO arguments
            "3LC/train_url": str(self.data.get("train").url),  # 3LC table used for training
            "3LC/val_url": val_url,  # 3LC table used for validation
            **{f"3LC/{k}": v for k, v in vars(self._settings).items()},  # 3LC settings
        }

        parameters = {
            key: value if not isinstance(value, Path) else value.as_posix() for key, value in parameters.items()
        }

        self._run.set_parameters(parameters)

    def _print_metrics_collection_epochs(self):
        """Print collection epochs to the console."""

        # Special message when no collection is enabled
        if self._settings.collection_disable:
            message = "No metrics collection is enabled."
        # No collection during training
        elif not self._metrics_collection_epochs:
            message = "Metrics will be collected after training only."
        # Print collection epochs
        else:
            if len(self._metrics_collection_epochs) == 1:
                epoch = str(next(iter(self._metrics_collection_epochs)))
                message = f"Metrics will be collected after training and after epoch {epoch}."
            else:
                epochs = ", ".join(str(epoch) for epoch in sorted(self._metrics_collection_epochs))
                message = f"Metrics will be collected after training and after the following epochs: {epochs}"

        LOGGER.info(f"{TLC_COLORSTR}{message}")

    def _print_task_specific_parameters(self):
        """Print task-specific parameters to the console."""

    def get_dataset(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def build_dataset(self, table, mode="train", batch=None):
        raise NotImplementedError("Subclasses must implement this method.")

    def get_validator(self, dataloader):
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def train_validator(self):
        """Validator for collecting 3LC metrics on the training set.

        In DDP mode, uses distributed dataloader so each GPU validates its portion.
        Metrics are gathered to RANK 0 in the validator's _post_validation.
        """
        if not self._train_validator:
            train_validator_dataloader = self.get_dataloader(
                self.data["train"],
                batch_size=self.batch_size if self.args.task == "obb" else self.batch_size * 2,
                rank=RANK,  # Distributed in DDP mode
                mode="val",
            )
            self._train_validator = self.get_validator(dataloader=train_validator_dataloader)
        return self._train_validator

    def validate(self):
        """Perform validation with 3LC metrics collection, also on the training data, if applicable.

        In DDP mode, train_validator uses distributed validation - each GPU validates its portion
        and metrics are gathered to RANK 0. The regular validation (super().validate()) also
        gathers 3LC metrics from all ranks.
        """
        # Validate on the training set, unless the training and validation sets are identical
        # All ranks participate in DDP mode for distributed metrics collection
        if (
            not self._settings.collection_disable
            and not self._settings.collection_val_only
            and self.epoch + 1 in self._metrics_collection_epochs
            and not self._train_equals_val()
        ):
            with _restore_random_state():
                self.train_validator(trainer=self)

        # Validate on the validation/test set like usual
        # In DDP mode, 3LC metrics are gathered from all ranks in the validator
        return super().validate()

    def _train_equals_val(self):
        if self._train_equals_val_result is not None:
            return self._train_equals_val_result

        self._train_equals_val_result = self.data["train"] is (self.data.get("val") or self.data["test"])
        if self._train_equals_val_result:
            LOGGER.info(
                f"{TLC_COLORSTR}Training and validation sets are identical. "
                "Skipping duplicate validation on the training set."
            )
        return self._train_equals_val_result

    def final_eval(self):
        """Perform final validation with metrics collection on both train and val sets.

        In DDP mode, uses distributed validation - each GPU validates its portion
        and metrics are gathered to RANK 0.
        """
        # Final validation on training set (all ranks participate in DDP mode)
        if not self._settings.collection_val_only and not self._settings.collection_disable:
            if self.best.exists() and not self._train_equals_val():
                with _restore_random_state():
                    self.train_validator._final_validation = True
                    self.train_validator._epoch = self.epoch
                    self.train_validator.data = self.data
                    self.train_validator(model=self.best)

        # Mark validator for final validation (all ranks, so gathering works correctly)
        if not self._settings.collection_disable:
            self.validator._final_validation = True

        super().final_eval()

        if RANK in {-1, 0}:
            self._save_confidence_metrics()
            if self._settings.image_embeddings_dim > 0:
                train_url = self.data["train"].url
                val_url = self.data["val"].url if "val" in self.data else self.data["test"].url
                foreign_table_url = train_url if not self._settings.collection_val_only else val_url
                reduce_embeddings(
                    self._run,
                    method=self._settings.image_embeddings_reducer,
                    n_components=self._settings.image_embeddings_dim,
                    foreign_table_url=foreign_table_url,
                    reducer_args=self._settings.image_embeddings_reducer_args,
                )
            self._run.set_status_completed()

    def _save_confidence_metrics(self):
        if self.args.task not in ("detect", "segment", "pose", "obb"):
            return

        try:
            # curves_results format: [[px, py_curve, x_label, y_label], ...]
            # Order: [PR, F1, Precision, Recall] (indices 0-3)
            box_curves = self.validator.metrics.box.curves_results
            px = box_curves[1][0]  # x values (confidence) from F1 curve

            curves = [box_curves[1][1], box_curves[3][1], box_curves[2][1]]  # F1, Recall, Precision
            names = ["F1_score", "Recall", "Precision"]

            if self.args.task == "pose":
                pose_curves = self.validator.metrics.pose.curves_results
                curves.extend([pose_curves[1][1], pose_curves[3][1], pose_curves[2][1]])
                names.extend(["Pose_F1_score", "Pose_Recall", "Pose_Precision"])

            if self.args.task == "segment":
                seg_curves = self.validator.metrics.seg.curves_results
                curves.extend([seg_curves[1][1], seg_curves[3][1], seg_curves[2][1]])
                names.extend(["Seg_F1_score", "Seg_Recall", "Seg_Precision"])

            values = {}
            for py, name in zip(curves, names):
                y = smooth(py.mean(0), 0.05)
                values[f"3LC/{name}"] = {"best_val": y.max(), "best_conf": px[y.argmax()]}

            self._run.set_parameters(values)
        except Exception as e:
            LOGGER.error(TLC_COLORSTR + f"Failed to save confidence metrics: {e}")

    def save_metrics(self, metrics):
        # Log aggregate metrics after every epoch
        processed_metrics = self._process_metrics(metrics)

        self._run.add_output_value({"epoch": self.epoch + 1, **processed_metrics})

        super().save_metrics(metrics=metrics)

    def _process_metrics(self, metrics):
        return metrics



from __future__ import annotations

import torch
import torch.nn.functional as F
from ultralytics.utils.loss import KeypointLoss, PoseLoss26, v8PoseLoss
from ultralytics.utils.ops import xyxy2xywh

from tlc_ultralytics.detect.loss import UnreducedBboxLoss


class v8UnreducedPoseLoss(v8PoseLoss):
    """v8PoseLoss that returns unreduced per-anchor losses for per-sample computation."""

    def __init__(self, model, training: bool = False):  # model must be de-paralleled
        super().__init__(model)
        # Prefer dataset-provided OKS sigmas if available on the model instance
        oks_sigmas = getattr(model, "oks_sigmas", None)
        if oks_sigmas is not None:
            sigmas_tensor = torch.as_tensor(oks_sigmas, device=self.device, dtype=torch.float32)
            self.keypoint_loss = KeypointLoss(sigmas=sigmas_tensor)
        # Replace bbox loss with unreduced variant and store training flag
        self.bbox_loss = UnreducedBboxLoss(self.reg_max)
        self.training = training

    def __call__(
        self,
        preds: dict[str, torch.Tensor] | tuple[torch.Tensor, dict[str, torch.Tensor]],
        batch: dict[str, torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        """Calculate unreduced losses for box, cls, dfl, pose (kpts), and kobj.

        Returns a dict of tensors shaped (batch, num_anchors) for each component.
        """
        preds_dict = preds[1] if isinstance(preds, (list, tuple)) else preds
        pred_distri = preds_dict["boxes"].permute(0, 2, 1).contiguous()
        pred_scores = preds_dict["scores"].permute(0, 2, 1).contiguous()
        feats = preds_dict["feats"]
        pred_kpts = preds_dict["kpts"].permute(0, 2, 1).contiguous()

        dtype = pred_scores.dtype
        batch_size = pred_scores.shape[0]
        imgsz = torch.tensor(feats[0].shape[2:], device=self.device, dtype=dtype) * self.stride[0]
        anchor_points, stride_tensor = self.make_anchors(feats)

        # Targets
        batch_idx = batch["batch_idx"].view(-1, 1)
        targets = torch.cat((batch_idx, batch["cls"].view(-1, 1), batch["bboxes"]), 1)
        targets = self.preprocess(targets.to(self.device), batch_size, scale_tensor=imgsz[[1, 0, 1, 0]])
        gt_labels, gt_bboxes = targets.split((1, 4), 2)  # cls, xyxy
        mask_gt = gt_bboxes.sum(2, keepdim=True).gt_(0.0)

        # Pboxes
        pred_bboxes = self.bbox_decode(anchor_points, pred_distri)  # (B, A, 4)
        pred_kpts = self.kpts_decode(anchor_points, pred_kpts.view(batch_size, -1, *self.kpt_shape))  # (B, A, K, D)

        # Assignment
        _, target_bboxes, target_scores, fg_mask, target_gt_idx = self.assigner(
            pred_scores.detach().sigmoid(),
            (pred_bboxes.detach() * stride_tensor).type(gt_bboxes.dtype),
            anchor_points * stride_tensor,
            gt_labels,
            gt_bboxes,
            mask_gt,
        )

        # Classification loss (per-anchor)
        cls_loss = self.bce(pred_scores, target_scores.to(dtype)).sum(dim=2)

        # Box and DFL losses (per-anchor, zeros where not fg)
        box_loss_full = torch.zeros_like(cls_loss)
        dfl_loss_full = torch.zeros_like(cls_loss)
        if fg_mask.sum():
            target_bboxes /= stride_tensor

            box_loss, dfl_loss = self.bbox_loss(
                pred_distri,
                pred_bboxes,
                anchor_points,
                target_bboxes,
                target_scores,
                max(target_scores.sum(), 1),
                fg_mask,
            )
            box_loss_full[fg_mask] = box_loss.to(cls_loss.dtype).squeeze()
            dfl_loss_full[fg_mask] = dfl_loss.to(cls_loss.dtype).squeeze()

        # Keypoints losses (pose and kobj), per-anchor on fg positions
        pose_loss_full = torch.zeros_like(cls_loss)
        kobj_loss_full = torch.zeros_like(cls_loss)
        if fg_mask.any():
            # Prepare selected GT keypoints batched per image as in v8PoseLoss.calculate_keypoints_loss
            keypoints = batch["keypoints"].to(self.device).float().clone()
            keypoints[..., 0] *= imgsz[1]
            keypoints[..., 1] *= imgsz[0]

            bidx = batch_idx.flatten()
            bs = len(fg_mask)
            max_kpts = torch.unique(bidx, return_counts=True)[1].max()
            batched_keypoints = torch.zeros(
                (bs, max_kpts, keypoints.shape[1], keypoints.shape[2]), device=keypoints.device
            )
            for i in range(bs):
                keypoints_i = keypoints[bidx == i]
                batched_keypoints[i, : keypoints_i.shape[0]] = keypoints_i

            target_gt_idx_expanded = target_gt_idx.unsqueeze(-1).unsqueeze(-1)
            selected_keypoints = batched_keypoints.gather(
                1, target_gt_idx_expanded.expand(-1, -1, keypoints.shape[1], keypoints.shape[2])
            )

            # Divide coords by stride to match pred scale
            selected_keypoints[..., :2] /= stride_tensor.view(1, -1, 1, 1)

            gt_kpt = selected_keypoints[fg_mask]
            area = xyxy2xywh(target_bboxes[fg_mask])[:, 2:].prod(1, keepdim=True)
            pred_kpt = pred_kpts[fg_mask]
            if gt_kpt.shape[-1] == 3:
                kpt_mask = gt_kpt[..., 2] != 0
            else:
                kpt_mask = torch.full_like(gt_kpt[..., 0], True, dtype=torch.bool)

            # Pose location loss per-anchor, replicate KeypointLoss but keep per-anchor reduction
            d = (pred_kpt[..., 0] - gt_kpt[..., 0]).pow(2) + (pred_kpt[..., 1] - gt_kpt[..., 1]).pow(2)
            # e = d / ((2 * self.keypoint_loss.sigmas) ** 2 * (area + 1e-9) * 2)
            e = d / ((2 * self.keypoint_loss.sigmas).pow(2) * (area + 1e-9) * 2)
            kpt_loss_factor = kpt_mask.shape[1] / (torch.sum(kpt_mask != 0, dim=1) + 1e-9)
            pose_loss_vec = (kpt_loss_factor.view(-1, 1) * ((1 - torch.exp(-e)) * kpt_mask)).mean(dim=1)

            # Keypoint objectness loss per-anchor
            if pred_kpt.shape[-1] == 3:
                kobj_loss_vec = F.binary_cross_entropy_with_logits(
                    pred_kpt[..., 2], kpt_mask.float(), reduction="none"
                ).mean(dim=1)
            else:
                kobj_loss_vec = torch.zeros_like(pose_loss_vec)

            # Scatter back into full tensors
            pose_loss_full[fg_mask] = pose_loss_vec.to(cls_loss.dtype)
            kobj_loss_full[fg_mask] = kobj_loss_vec.to(cls_loss.dtype)

        losses: dict[str, torch.Tensor] = {
            "cls_loss": cls_loss,
            "box_loss": box_loss_full,
            "dfl_loss": dfl_loss_full,
            "pose_loss": pose_loss_full,
            "kobj_loss": kobj_loss_full,
        }

        if self.training:
            cls_weight = self.hyp.cls if hasattr(self.hyp, "cls") else self.hyp["cls"]
            box_weight = self.hyp.box if hasattr(self.hyp, "box") else self.hyp["box"]
            dfl_weight = self.hyp.dfl if hasattr(self.hyp, "dfl") else self.hyp["dfl"]
            pose_weight = self.hyp.pose if hasattr(self.hyp, "pose") else self.hyp["pose"]
            kobj_weight = self.hyp.kobj if hasattr(self.hyp, "kobj") else self.hyp["kobj"]
            losses["loss"] = (
                cls_weight * cls_loss
                + box_weight * box_loss_full
                + dfl_weight * dfl_loss_full
                + pose_weight * pose_loss_full
                + kobj_weight * kobj_loss_full
            )

        return losses

    def make_anchors(self, feats):
        # Helper to avoid importing at top-level to keep parity with ultralytics.utils.loss
        from ultralytics.utils.tal import make_anchors

        anchor_points, stride_tensor = make_anchors(feats, self.stride, 0.5)
        return anchor_points, stride_tensor


class TLCv8PoseLoss(v8PoseLoss):
    """Pose loss that prefers dataset-provided OKS sigmas when available.

    If the attached model has an attribute `oks_sigmas` (list/ndarray/tensor), use it to
    configure the `KeypointLoss`. Falls back to the default behavior otherwise.
    """

    def __init__(self, model, tal_topk: int = 10, tal_topk2: int | None = None):
        super().__init__(model, tal_topk, tal_topk2)
        oks_sigmas = getattr(model, "oks_sigmas", None)
        if oks_sigmas is not None:
            sigmas_tensor = torch.as_tensor(oks_sigmas, device=self.device, dtype=torch.float32)
            self.keypoint_loss = KeypointLoss(sigmas=sigmas_tensor)


class TLCPoseLoss26(PoseLoss26):
    """PoseLoss26 that prefers dataset-provided OKS sigmas when available.

    If the attached model has an attribute `oks_sigmas` (list/ndarray/tensor), use it to
    configure the `KeypointLoss`. Falls back to the default behavior otherwise.
    """

    def __init__(self, model, tal_topk: int = 10, tal_topk2: int | None = None):
        super().__init__(model, tal_topk, tal_topk2)
        oks_sigmas = getattr(model, "oks_sigmas", None)
        if oks_sigmas is not None:
            sigmas_tensor = torch.as_tensor(oks_sigmas, device=self.device, dtype=torch.float32)
            self.keypoint_loss = KeypointLoss(sigmas=sigmas_tensor)



from __future__ import annotations

from typing import Any

import numpy as np
import tlc
import torch
from tlc.core.builtins.constants import KEYPOINTS_2D_PREDICTED
from tlc.core.builtins.schemas import Keypoints2DSchema
from tlc.core.data_formats import Keypoints2DInstances
from ultralytics.models.yolo.pose.val import PoseValidator
from ultralytics.utils import LOGGER, ops

from tlc_ultralytics.constants import IMAGE_COLUMN_NAME, POSE_LABEL_COLUMN_NAME, TLC_COLORSTR
from tlc_ultralytics.engine.validator import TLCValidatorMixin
from tlc_ultralytics.pose.dataset import TLCYOLOPoseDataset
from tlc_ultralytics.pose.loss import v8UnreducedPoseLoss
from tlc_ultralytics.pose.utils import yolo_pose_loss_schemas
from tlc_ultralytics.utils.dataset import check_tlc_dataset


class TLCPoseValidator(TLCValidatorMixin, PoseValidator):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = POSE_LABEL_COLUMN_NAME

    def check_dataset(self, *args, **kwargs):
        return check_tlc_dataset(*args, task="pose", settings=self._settings, **kwargs)

    def build_dataset(self, table, mode: str = "val", batch=None):
        return TLCYOLOPoseDataset(
            table,
            data=self.data,
            exclude_zero=self._settings.exclude_zero_weight_collection,
            class_map=self.data["3lc_class_to_range"],
            imgsz=self.args.imgsz,
            batch_size=batch,
            augment=mode == "train",
            rect=self.args.rect,
            cache=self.args.cache or None,
            single_cls=self.args.single_cls or False,
            stride=int(self.stride),
            pad=0.5,
            prefix=self.args.split or mode,
            task=self.args.task,
            classes=self.args.classes,
            fraction=1.0,
            image_column_name=self._settings.image_column_name,
            label_column_name=self._settings.label_column_name,
        )

    def init_metrics(self, model: torch.nn.Module) -> None:
        super().init_metrics(model)

        if self.data.get("oks_sigmas"):
            self.sigma = np.array(self.data.get("oks_sigmas"))

    def postprocess(self, preds):
        self._curr_raw_preds = preds if self._settings.collect_loss else None
        return super().postprocess(preds)

    def _get_metrics_schemas(self) -> dict[str, tlc.Schema]:
        predicted_pose_schema = Keypoints2DSchema(
            classes=self.data["names"],
            num_keypoints=self.kpt_shape[0],
            points=self.data.get("points"),
            point_attributes=self.data.get("keypoint_attributes"),
            lines=self.data.get("lines"),
            line_attributes=self.data.get("line_attributes"),
            triangles=self.data.get("triangles"),
            triangle_attributes=self.data.get("triangle_attributes"),
            include_per_instance_confidence=True,
            include_per_point_confidence=self.kpt_shape[1] == 3,
            writable=False,
        )

        loss_schemas = yolo_pose_loss_schemas(training=self._training) if self._settings.collect_loss else {}
        return {KEYPOINTS_2D_PREDICTED: predicted_pose_schema, **loss_schemas}

    def _compute_3lc_metrics(self, preds, batch) -> dict[str, Any]:
        predicted = []

        for i, pred in enumerate(preds):
            (
                predicted_keypoints,
                predicted_confidences,
                predicted_classes,
                predicted_bboxes,
            ) = (
                pred["keypoints"].clone(),
                pred["conf"].clone(),
                pred["cls"].clone(),
                pred["bboxes"].clone(),
            )
            h, w = batch["ori_shape"][i]

            builder = Keypoints2DInstances.create_empty(
                image_height=int(h),
                image_width=int(w),
                include_instance_confidences=True,
            )

            # Filter out low confidence predictions
            mask = predicted_confidences > self._settings.conf_thres
            if not mask.any():
                predicted.append(builder.to_row())
                continue

            predicted_keypoints = predicted_keypoints[mask]
            predicted_confidences = predicted_confidences[mask].tolist()
            predicted_classes = predicted_classes[mask].cpu().numpy().astype(np.int32).tolist()
            predicted_bboxes = predicted_bboxes[mask]

            resized_shape = batch["resized_shape"][i]
            ori_shape = batch["ori_shape"][i]
            ratio_pad = batch["ratio_pad"][i]
            scaled_bboxes = ops.scale_boxes(resized_shape, predicted_bboxes, ori_shape, ratio_pad)
            scaled_keypoints = ops.scale_coords(resized_shape, predicted_keypoints, ori_shape, ratio_pad)

            for j in range(len(predicted_keypoints)):
                predicted_kpts = scaled_keypoints[j]
                predicted_bbox = scaled_bboxes[j].cpu().numpy().astype(np.float32)
                kxy = predicted_kpts[:, 0:2].cpu().numpy().astype(np.float32)
                kconf = (
                    predicted_kpts[:, 2].cpu().numpy().astype(np.float32).tolist()
                    if predicted_kpts.shape[1] == 3
                    else None
                )
                builder.add_instance(
                    keypoints=kxy,
                    bbox=predicted_bbox.tolist(),
                    label=int(predicted_classes[j]),
                    confidence=kconf,
                    normalized=False,
                    bbox_format="xyxy",
                    instance_confidence=float(predicted_confidences[j]),
                )

            row = builder.to_row()
            predicted.append(row)

        losses = self.loss_fn(self._curr_raw_preds, batch) if self._settings.collect_loss else {}
        return {
            KEYPOINTS_2D_PREDICTED: predicted,
            **{k: tensor.mean(dim=1).cpu().numpy() for k, tensor in losses.items()},
        }

    def _prepare_loss_fn(self, model):
        loss_model = model.model if hasattr(model.model, "model") else model

        # Check if this is a YOLO26 (end2end) model - per-sample loss is not supported for these
        is_end2end = getattr(loss_model.model[-1], "end2end", False) if hasattr(loss_model, "model") else False

        if is_end2end and self._settings.collect_loss:
            LOGGER.warning(
                f"{TLC_COLORSTR}Per-sample loss collection is not supported for YOLO26 (end2end) models. "
                "Disabling loss collection for this run."
            )
            self._settings.collect_loss = False
            return

        if self._settings.collect_loss:
            # Pass through dataset-provided OKS sigmas to the loss via model attribute for consistency
            oks_sigmas = self._settings.oks_sigmas or self.data.get("oks_sigmas")
            if oks_sigmas is not None:
                loss_model.oks_sigmas = oks_sigmas
            self.loss_fn = v8UnreducedPoseLoss(loss_model, training=self._training)

    def _add_embeddings_hook(self, model) -> int:
        if hasattr(model.model, "model"):
            model = model.model

        sppf_index = next((i for i, m in enumerate(model.model) if "SPPF" in m.type), -1)
        if sppf_index == -1:
            return 0

        weak_self = self

        def hook_fn(_module, _input, output):
            flat = torch.nn.functional.adaptive_avg_pool2d(output, (1, 1)).squeeze(-1).squeeze(-1)
            weak_self.embeddings = flat.detach().cpu().numpy()

        self._hook_handles.append(model.model[sppf_index].register_forward_hook(hook_fn))
        return model.model[sppf_index]._modules["cv2"]._modules["conv"].out_channels

    def _infer_batch_size(self, preds, batch) -> int:
        return len(batch["im_file"])



from ultralytics.utils.dist import generate_ddp_command as ultralytics_generate_ddp_command


def generate_ddp_command(trainer):
    """Override the generate_ddp_command function to serialize the 3LC data required for DDP processes
    as part of the data argument. The override is temporary and therefore does not alter the original trainer state.
    """
    original_data = trainer.args.data
    trainer.args.data = trainer._serialize_state()
    command = ultralytics_generate_ddp_command(trainer)
    trainer.args.data = original_data
    return command



from __future__ import annotations

import numpy as np
import tlc
import torch.distributed as dist
import ultralytics
from ultralytics.engine.validator import BaseValidator
from ultralytics.utils import LOGGER, RANK, colorstr

from tlc_ultralytics.constants import (
    DEFAULT_COLLECT_RUN_DESCRIPTION,
    MAP,
    MAP50_95,
    MAP50_95_SEG,
    MAP_SEG,
    NUM_IMAGES,
    NUM_INSTANCES,
    PER_CLASS_METRICS_STREAM_NAME,
    PRECISION,
    PRECISION_SEG,
    RECALL,
    RECALL_SEG,
    TLC_COLORSTR,
    TRAINING_PHASE,
)
from tlc_ultralytics.engine.utils import _complete_label_column_name, _handle_deprecated_column_name
from tlc_ultralytics.settings import Settings
from tlc_ultralytics.utils import image_embeddings_schema, training_phase_schema


def execute_when_collecting(method):
    def wrapper(self, *args, **kwargs):
        if self._should_collect:
            return method(self, *args, **kwargs)

    return wrapper


class TLCValidatorMixin(BaseValidator):
    def __init__(
        self,
        *args,
        run: tlc.Run | None = None,
        image_column_name: str | None = None,
        label_column_name: str | None = None,
        settings: Settings | None = None,
        training: bool = False,
        **kwargs,
    ):
        self._run = run
        # Settings can be passed as an argument directly to the validator, or as a keyword from the trainer
        self._settings = settings or kwargs.get("args", {}).pop("settings", None) or Settings()

        self._settings.image_column_name = _handle_deprecated_column_name(
            image_column_name,
            self._settings.image_column_name,
            self._default_image_column_name,
            column_name="image_column_name",
        )
        self._settings.label_column_name = _handle_deprecated_column_name(
            label_column_name,
            self._settings.label_column_name,
            self._default_label_column_name,
            column_name="label_column_name",
        )
        self._settings.label_column_name = _complete_label_column_name(
            self._settings.label_column_name,
            self._default_label_column_name,
        )

        self._training = training

        if not training:
            self._settings.verify(training=False)

        # Table is passed when doing validation only, not when training
        self._table = kwargs.get("args", {}).pop("table", None) if not training else None

        # State
        self._epoch = None
        self._should_collect = None
        self._seen = None
        self._final_validation = False
        self._hook_handles = []

        super().__init__(*args, **kwargs)

        if not self._training:
            # Set up dataset checking bypass before calling check_dataset
            self.data = self.check_dataset(
                self.args.data,
                {self.args.split: self._table} if self._table is not None else None,
                self._settings.image_column_name,
                self._settings.label_column_name,
                project_name=self._settings.project_name,
                splits=(self.args.split,),
            )

        # Create a run if not provided
        if self._run is None:
            # Reuse active run only if it has the same project name (if a different run is active)
            if self._settings.project_name:
                project_name = self._settings.project_name
            else:
                first_split = next(iter(self.data.keys()))
                project_name = self.data[first_split].project_name
                LOGGER.info(f"{TLC_COLORSTR}Using project name '{project_name}' from the provided table to create run.")

            if tlc.active_run() and tlc.active_run().project_name == project_name:
                self._run = tlc.active_run()
                run_name = self._run.url.parts[-1]
                LOGGER.info(f"{TLC_COLORSTR}Using active run named '{run_name}' in project {self._run.project_name}.")
            else:
                try:
                    root_url = self.data[first_split].root
                except Exception:
                    root_url = None

                self._run = tlc.init(
                    project_name=project_name,
                    description=self._settings.run_description
                    if self._settings.run_description
                    else DEFAULT_COLLECT_RUN_DESCRIPTION,
                    run_name=self._settings.run_name,
                    root_url=root_url,
                )
                LOGGER.info(
                    f"{TLC_COLORSTR}Created run named '{self._run.url.parts[-1]}' in project {self._run.project_name}."
                )

        if self.args.task == "pose" and not self._training:
            table_sigmas = self.data.get("oks_sigmas")
            if table_sigmas is not None:
                table_sigmas_rounded = [round(x, 2) for x in table_sigmas]
                LOGGER.info(f"{TLC_COLORSTR}Using OKS sigmas: {table_sigmas_rounded} from Table for validation")

        self.metrics.run_url = self._run.url

    def __call__(self, trainer=None, model=None):
        self._epoch = trainer.epoch if trainer is not None else self._epoch

        if trainer:
            self._should_collect = (
                not self._settings.collection_disable and self._epoch + 1 in trainer._metrics_collection_epochs
            )
        else:
            self._should_collect = not self._settings.collection_disable

        # Define bypass functions that use our data object
        def bypass_check_det_dataset(*args, **kwargs):
            return self.data

        def bypass_check_cls_dataset(*args, **kwargs):
            return self.data

        # Patch the functions in ultralytics.data.utils on the validator module
        ultralytics.engine.validator.check_det_dataset = bypass_check_det_dataset
        ultralytics.engine.validator.check_cls_dataset = bypass_check_cls_dataset

        try:
            # Call parent to perform the validation
            out = super().__call__(trainer, model)
        finally:
            # Restore original functions
            ultralytics.engine.validator.check_det_dataset = ultralytics.data.utils.check_det_dataset
            ultralytics.engine.validator.check_cls_dataset = ultralytics.data.utils.check_cls_dataset

        # Per-class metrics only on RANK 0 (uses aggregate metrics from gather_stats)
        if RANK in {-1, 0}:
            self._write_per_class_metrics_tables()

        # All ranks call _post_validation to participate in distributed gathering
        # (the method handles RANK checks internally for the actual run updates)
        self._post_validation()

        return out

    def get_desc(self):
        """Add the split name next to the validation description"""
        desc = super().get_desc()

        split = self.dataloader.dataset.display_name.split("-")[-1]  # get final part
        initial_spaces = len(desc) - len(desc.lstrip())
        split_centered = split.center(initial_spaces)
        split_str = f"{colorstr(split_centered)}"
        desc = split_str + desc[len(split_centered) :]

        return desc

    def init_metrics(self, model):
        super().init_metrics(model)

        self._verify_model_data_compatibility(model.names)
        self._pre_validation(model)

    def build_dataset(self, table):
        """Build a dataset from a table"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _verify_model_data_compatibility(self, names):
        """Verify that the model being validated is compatible with the data"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _get_metrics_schemas(self) -> dict[str, tlc.Schema]:
        """Get the metrics schemas for the 3LC metrics data"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _compute_3lc_metrics(self, preds, batch) -> dict[str, tlc.MetricData]:
        """Compute 3LC metrics for a batch of predictions and targets"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _add_embeddings_hook(self, model) -> int:
        """Add a hook to extract embeddings from the model, and infer the activation size"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _infer_batch_size(self, preds) -> int:
        """Infer the batch size from the predictions"""
        raise NotImplementedError("Subclasses must implement this method.")

    def _prepare_loss_fn(self, model):
        pass

    def update_metrics(self, preds, batch):
        """Collect 3LC metrics"""
        self._update_metrics(preds, batch)

        # Let parent collect its own metrics
        super().update_metrics(preds, batch)

    @execute_when_collecting
    def _update_metrics(self, preds, batch):
        """Update 3LC metrics with common and task-specific metrics.

        In DDP mode, each rank collects metrics for its portion of the data.
        These are gathered to RANK 0 in _post_validation.
        """
        batch_size = self._infer_batch_size(preds, batch)

        batch_metrics = {
            tlc.EXAMPLE_ID: [int(example_id) for example_id in batch["example_id"]],
            **self._compute_3lc_metrics(preds, batch),  # Task specific metrics
        }

        if self._settings.metrics_collection_function:
            batch_metrics.update(self._settings.metrics_collection_function(preds, batch))

        if self._settings.image_embeddings_dim > 0:
            batch_metrics["embeddings"] = self.embeddings

        if self._training:
            batch_metrics[tlc.EPOCH] = [self._epoch + 1] * batch_size
            training_phase = 1 if self._final_validation else 0
            batch_metrics[TRAINING_PHASE] = [training_phase] * batch_size

        # Add DDP rank for distributed validation debugging
        if RANK >= 0:
            batch_metrics["ddp_rank"] = [RANK] * batch_size

        self._metrics_writer.add_batch(batch_metrics)
        self._seen += batch_size

    @execute_when_collecting
    def _pre_validation(self, model):
        """Prepare the validator for metrics collection.

        In DDP mode, each rank prepares its own metrics writer.
        """
        column_schemas = {}
        column_schemas.update(self._get_metrics_schemas())  # Add task-specific metrics schema

        if self._settings.metrics_schemas:
            column_schemas.update(self._settings.metrics_schemas)

        self._prepare_loss_fn(model)

        if self._settings.image_embeddings_dim > 0:
            # Add hook and get the activation size
            activation_size = self._add_embeddings_hook(model)

            column_schemas["embeddings"] = image_embeddings_schema(activation_size=activation_size)

        if self._epoch is not None:
            column_schemas[TRAINING_PHASE] = training_phase_schema()

        # Add DDP rank column for distributed validation debugging
        if RANK >= 0:
            column_schemas["ddp_rank"] = tlc.Schema(
                display_name="DDP rank",
                value=tlc.Int32Value(),
                description="DDP rank that processed this sample",
                default_visible=False,
            )

        # Only RANK 0 (or single GPU) updates run status
        if RANK in {-1, 0}:
            self._run.set_status_collecting()

        self._metrics_writer = tlc.MetricsTableWriter(
            run_url=self._run.url,
            foreign_table_url=self.dataloader.dataset.table.url,
            column_schemas=column_schemas,
        )

        self._seen = 0

    @execute_when_collecting
    def _post_validation(self):
        """Clean up the validator after one validation pass.

        In DDP mode, gathers metrics_infos and input table URLs from all ranks
        to RANK 0, which then updates the run with all collected data.
        """
        # Each rank finalizes its own metrics writer
        self._metrics_writer.finalize()
        metrics_infos = self._metrics_writer.get_written_metrics_infos()
        input_table_url = self.dataloader.dataset.table.url.to_str()

        # Gather metrics from all ranks to RANK 0 in DDP mode
        if RANK >= 0:
            world_size = dist.get_world_size()
            gathered_metrics_infos = [None] * world_size if RANK == 0 else None
            gathered_input_urls = [None] * world_size if RANK == 0 else None

            dist.gather_object(metrics_infos, gathered_metrics_infos, dst=0)
            dist.gather_object(input_table_url, gathered_input_urls, dst=0)

            if RANK == 0:
                # Flatten metrics_infos from all ranks
                all_metrics_infos = []
                for rank_metrics in gathered_metrics_infos:
                    all_metrics_infos.extend(rank_metrics)
                metrics_infos = all_metrics_infos

                # Collect unique input table URLs (should all be the same in distributed validation)
                input_table_urls = list(set(gathered_input_urls))
        else:
            # Single GPU mode (RANK == -1)
            input_table_urls = [input_table_url]

        # Only RANK 0 (or single GPU) updates the run
        if RANK in {-1, 0}:
            self._run.update_metrics(metrics_infos)

            for url in input_table_urls:
                self._run.add_input_table(tlc.Url(url))

            # Improve memory usage - don't cache metrics data
            for metrics_info in metrics_infos:
                tlc.ObjectRegistry._delete_object_from_caches(tlc.Url(metrics_info["url"]).to_absolute(self._run.url))

            self._run.set_status_running()

        # Remove hook handles (all ranks)
        if self._settings.image_embeddings_dim > 0:
            for handle in self._hook_handles:
                handle.remove()
            self._hook_handles.clear()

        # Reset state (all ranks)
        self._seen = None
        self._training_phase = None
        self._final_validation = None

    def _write_per_class_metrics_tables(self) -> None:
        if self.args.task not in ("detect", "segment", "obb"):
            # Per-class metrics currently only supported for detection, segmentation, and obb tasks
            return

        metrics_writer = tlc.MetricsTableWriter(
            run_url=self._run.url,
            column_schemas=self._per_class_metrics_schemas(),
        )

        epoch = self._epoch + 1 if self._epoch is not None else -1
        training_phase = 1 if self._final_validation else 0
        num_classes = self.nc + 1  # all classes plus "all"

        metrics_batch = (
            {
                tlc.EPOCH: [epoch] * num_classes,
                TRAINING_PHASE: [training_phase] * num_classes,
            }
            if self._training
            else {}
        )

        metrics_batch.update(
            {
                tlc.FOREIGN_TABLE_ID: [0] * num_classes,
                tlc.LABEL: list(range(num_classes)),
                NUM_INSTANCES: np.append(self.metrics.nt_per_class, self.metrics.nt_per_class.sum()),
                NUM_IMAGES: np.append(self.metrics.nt_per_image, self.seen),
                **self._generate_per_class_metrics(),
            }
        )

        metrics_writer.add_batch(metrics_batch)
        metrics_writer.finalize()
        metrics_infos = metrics_writer.get_written_metrics_infos()
        for m in metrics_infos:
            # Set the stream name to the per-class metrics stream
            # TODO: This should be a constructor argument to MetricsTableWriter
            m["stream_name"] = PER_CLASS_METRICS_STREAM_NAME
        self._run.update_metrics(metrics_infos)

    def _per_class_metrics_schemas(self):
        metrics_schemas = {
            TRAINING_PHASE: training_phase_schema(),
            tlc.FOREIGN_TABLE_ID: tlc.ForeignTableIdSchema(
                self.dataloader.dataset.table.url.to_relative(self._run.url / "metrics").to_str(),
            ),
            tlc.LABEL: tlc.CategoricalLabel("class", {**self.names, self.nc: "all"}).schema,
            NUM_IMAGES: tlc.Schema(
                value=tlc.Int32Value(),
                description="Number of images with at least one instance of the class",
            ),
            NUM_INSTANCES: tlc.Schema(
                value=tlc.Int32Value(),
                description="Total number of instances of the class in all images",
            ),
            PRECISION: tlc.Schema(
                value=tlc.Float32Value(),
                description="Precision of the class",
            ),
            RECALL: tlc.Schema(
                value=tlc.Float32Value(),
                description="Recall of the class",
            ),
            MAP: tlc.Schema(
                value=tlc.Float32Value(),
                description="mAP of the class",
            ),
            MAP50_95: tlc.Schema(
                value=tlc.Float32Value(),
                description="mAP50-95 of the class",
            ),
        }

        if self.args.task == "segment":
            metrics_schemas[PRECISION_SEG] = tlc.Schema(
                value=tlc.Float32Value(),
                description="Mask precision of the class",
            )

            metrics_schemas[RECALL_SEG] = tlc.Schema(
                value=tlc.Float32Value(),
                description="Mask recall of the class",
            )

            metrics_schemas[MAP_SEG] = tlc.Schema(
                value=tlc.Float32Value(),
                description="Mask mAP of the class",
            )

            metrics_schemas[MAP50_95_SEG] = tlc.Schema(
                value=tlc.Float32Value(),
                description="Mask mAP50-95 of the class",
            )
        return metrics_schemas

    def _generate_per_class_metrics(self):
        """Transform metrics from self.metrics to a format suitable for 3LC"""
        # Consider moving this to TLCDetectionValidator when supporting other tasks
        precisions = np.zeros(self.nc + 1)
        recalls = np.zeros(self.nc + 1)
        mAPs = np.zeros(self.nc + 1)
        mAP50_95s = np.zeros(self.nc + 1)

        if self.args.task == "segment":
            precisions_seg = np.zeros(self.nc + 1)
            recalls_seg = np.zeros(self.nc + 1)
            mAPs_seg = np.zeros(self.nc + 1)
            mAP50_95s_seg = np.zeros(self.nc + 1)

        for i in range(self.nc):
            if i in self.metrics.ap_class_index:
                class_results = self.metrics.class_result(np.where(self.metrics.ap_class_index == i)[0][0])
                if self.args.task in ("detect", "obb"):
                    p, r, ap50, ap5095 = class_results
                else:  # seg and pose
                    p, r, ap50, ap5095, p_seg, r_seg, ap50_seg, ap5095_seg = class_results
            else:
                p, r, ap50, ap5095 = 0.0, 0.0, 0.0, 0.0
                if self.args.task == "segment":
                    p_seg, r_seg, ap50_seg, ap5095_seg = 0.0, 0.0, 0.0, 0.0

            precisions[i] = p
            recalls[i] = r
            mAPs[i] = ap50
            mAP50_95s[i] = ap5095

            if self.args.task == "segment":
                precisions_seg[i] = p_seg
                recalls_seg[i] = r_seg
                mAPs_seg[i] = ap50_seg
                mAP50_95s_seg[i] = ap5095_seg

        mean_results = self.metrics.mean_results()
        if self.args.task in ("detect", "obb"):
            all_p, all_r, all_mAP50, all_mAP50_95 = mean_results
        else:  # seg and pose
            (
                all_p,
                all_r,
                all_mAP50,
                all_mAP50_95,
                all_p_seg,
                all_r_seg,
                all_mAP50_seg,
                all_mAP50_95_seg,
            ) = mean_results

        precisions[self.nc] = all_p
        recalls[self.nc] = all_r
        mAPs[self.nc] = all_mAP50
        mAP50_95s[self.nc] = all_mAP50_95

        metrics = {
            PRECISION: precisions,
            RECALL: recalls,
            MAP: mAPs,
            MAP50_95: mAP50_95s,
        }

        if self.args.task == "segment":
            precisions_seg[self.nc] = all_p_seg
            recalls_seg[self.nc] = all_r_seg
            mAPs_seg[self.nc] = all_mAP50_seg
            mAP50_95s_seg[self.nc] = all_mAP50_95_seg

            metrics[PRECISION_SEG] = precisions_seg
            metrics[RECALL_SEG] = recalls_seg
            metrics[MAP_SEG] = mAPs_seg
            metrics[MAP50_95_SEG] = mAP50_95s_seg

        return metrics

    def _verify_model_data_compatibility(self, model_class_names):
        """Verify that the model classes match the dataset classes. For a classification model, this amounts to checking
        that the order of the class names match and that they have the same number of classes."""
        dataset_class_names = self.data["names"]

        if len(model_class_names) != len(dataset_class_names):
            raise ValueError(
                f"The model and data are incompatible. The model was trained on {len(model_class_names)} classes, "
                f"but the data has {len(dataset_class_names)} classes. "
            )

        # Imagenet has a class name transform in YOLO which is not applied on table creation.
        # TODO: Remove when image_folder takes a sparse class name mapping to change these
        if "n01440764" not in set(dataset_class_names.values()):
            if model_class_names != dataset_class_names:
                raise ValueError(
                    "The model was trained on a different set of classes to the classes in the dataset, "
                    "or the classes are in a different order."
                )



from copy import deepcopy

from ultralytics.models.yolo.obb.train import OBBTrainer

from tlc_ultralytics.constants import OBB_LABEL_COLUMN_NAME
from tlc_ultralytics.detect.trainer import TLCDetectionTrainer
from tlc_ultralytics.obb.validator import TLCOBBValidator
from tlc_ultralytics.utils.dataset import check_tlc_dataset


class TLCOBBTrainer(OBBTrainer, TLCDetectionTrainer):
    _default_label_column_name = OBB_LABEL_COLUMN_NAME

    def get_dataset(self):
        # Parse yaml and create tables
        self.data = check_tlc_dataset(
            self.args.data,
            self._tables,
            self._settings.image_column_name,
            self._settings.label_column_name,
            project_name=self._settings.project_name,
            splits=("train", "val"),
            task="obb",
            settings=self._settings,
        )

        # Get test data if val not present
        if "val" not in self.data:
            data_test = check_tlc_dataset(
                self.args.data,
                self._tables,
                self._settings.image_column_name,
                self._settings.label_column_name,
                project_name=self._settings.project_name,
                splits=("test",),
                task="obb",
                settings=self._settings,
            )
            self.data["test"] = data_test["test"]

        return self.data

    def _process_metrics(self, metrics):
        detection_metrics = super()._process_metrics(metrics)
        return detection_metrics

    def get_validator(self, dataloader=None):
        self.loss_names = "box_loss", "seg_loss", "cls_loss", "dfl_loss"

        if not dataloader:
            dataloader = self.test_loader

        return TLCOBBValidator(
            dataloader,
            save_dir=self.save_dir,
            args=deepcopy(self.args),
            _callbacks=self.callbacks,
            run=self._run,
            settings=self._settings,
            training=True,
        )



from __future__ import annotations

import importlib
from importlib.metadata import PackageNotFoundError, metadata

from packaging.specifiers import SpecifierSet
from ultralytics.utils import LOGGER

from tlc_ultralytics.constants import REQUIREMENTS_TO_CHECK


def check_requirements(requirements_to_check: list[tuple[str, str]] = REQUIREMENTS_TO_CHECK) -> None:
    """Check the versions of the required packages are installed, and warn if the versions are not
    known to be compatible.

    :param requirements_to_check: List of tuples of (package name, import name) of packages to check.
    :raises ImportError: If the requirements are not installed.
    """
    try:
        tlc_ultralytics_requirements = metadata("3lc-ultralytics").json["requires_dist"]
    except PackageNotFoundError:
        msg = (
            "No version specifier found for '3lc-ultralytics', likely due to the integration not being installed. It "
            "is recommended to install the integration from PyPI with `pip install 3lc-ultralytics` or equivalent, "
            "or `pip install -e .` if installing from source."
        )
        LOGGER.warning(msg)
        return

    for package_name, import_name in requirements_to_check:
        try:
            importlib.import_module(import_name)
        except ImportError:
            msg = (
                f"Failed to import '{import_name}', which is a required dependency of 3lc-ultralytics. "
                f"Please install it with `pip install {package_name}` or equivalent."
            )
            raise ImportError(msg) from None

        # Check the version of the package and match it against the version set specifier
        installed_version = importlib.metadata.version(package_name)
        required_version_specifier = next(
            version_specifier.replace(package_name, "")
            for version_specifier in tlc_ultralytics_requirements
            if version_specifier.startswith(package_name)
        )

        if required_version_specifier is None:
            LOGGER.info(f"No version specifier found for '{package_name}', skipping version check.")

        if installed_version not in SpecifierSet(required_version_specifier):
            LOGGER.warning(
                f"The installed version of '{package_name}' ({installed_version}) is outside the required version "
                f"range {required_version_specifier}. This may cause compatibility issues, please use a compatible "
                "version if issues are encountered."
            )



from __future__ import annotations

from functools import partial
from typing import Any

import tlc
from ultralytics.data.dataset import ClassificationDataset

from tlc_ultralytics.engine.dataset import TLCDatasetMixin


class _DummyImageFolder:
    def __init__(self, root: tlc.Table, allow_empty: bool = False, samples: list[tuple[str, int]] | None = None):
        self._root = root
        self._samples = samples

    @property
    def samples(self):
        return self._samples

    @property
    def root(self):
        return self._root.url


class TLCClassificationDataset(TLCDatasetMixin, ClassificationDataset):
    """
    Initialize 3LC classification dataset for use in YOLO classification.

    Args:
        table (tlc.Table): The 3LC table with classification data. Needs columns 'image' and 'label'.
        args (Namespace): See parent.
        augment (bool): See parent.
        prefix (str): See parent.

    """

    def __init__(
        self,
        table,
        args,
        augment=False,
        prefix="",
        image_column_name=tlc.IMAGE,
        label_column_name=tlc.LABEL,
        exclude_zero=False,
        class_map=None,
    ):
        # Populate self.samples with image paths and labels
        # Each is a tuple of (image_path, label)
        assert isinstance(table, tlc.Table)
        self.table = table
        self.root = table.url
        self.prefix = prefix
        self._image_column_name = image_column_name
        self._label_column_name = label_column_name
        self._exclude_zero = exclude_zero
        self._class_map = class_map
        self._example_ids = []

        self.verify_schema()

        im_files, labels = self._get_rows_from_table()
        self.samples = list(zip(im_files, labels))

        # Override torchvision ImageFolder when called by parent __init__
        import torchvision

        OriginalImageFolder = torchvision.datasets.ImageFolder
        torchvision.datasets.ImageFolder = partial(_DummyImageFolder, samples=self.samples)

        ClassificationDataset.__init__(self, table, args, augment=augment, prefix=prefix)

        # Restore torchvision ImageFolder
        torchvision.datasets.ImageFolder = OriginalImageFolder

        # Call mixin
        self._post_init()

    def verify_schema(self):
        """Verify that the provided Table has the desired entries"""

        # Check for data in columns
        assert len(self.table) > 0, f"Table {self.root.to_str()} has no rows."
        first_row = self.table.table_rows[0]
        assert isinstance(first_row[self._image_column_name], str), (
            f"First value in image column '{self._image_column_name}' in table {self.root.to_str()} is not a string."
        )
        assert isinstance(first_row[self._label_column_name], int), (
            f"First value in label column '{self._label_column_name}' in table {self.root.to_str()} is not an integer."
        )

    def verify_images(self):
        """Called by parent init_attributes, but this is handled by the 3LC mixin."""
        return self.samples

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> Any:
        label = row[self._label_column_name]

        if self._class_map:
            label = self._class_map[label]

        self._example_ids.append(example_id)

        return label

    def _index_to_example_id(self, index: int) -> int:
        """Get the example id for the given index."""
        return self._example_ids[index]



from typing import Optional

import ultralytics
from torch.utils.data import Sampler
from ultralytics.data.build import InfiniteDataLoader
from ultralytics.data.build import build_dataloader as build_dataloader_ultralytics


def build_dataloader(*args, **kwargs):
    sampler: Optional[Sampler] = kwargs.pop("sampler", None)

    class _InfiniteDataLoaderWithSampler(InfiniteDataLoader):
        """This class is a temporary patch to the InfiniteDataLoader that allows for a custom sampler to be set.

        It is necessary because the `build_dataloader` function in Ultralytics does not allow for a custom sampler to be
        set.
        """

        def __init__(self, *args, **kwargs):
            nonlocal sampler

            provided_sampler = kwargs.pop("sampler", None)
            if provided_sampler and sampler:
                msg = "Cannot patch InfiniteDataLoader when a sampler is already set by Ultralytics. "
                msg += "The 3LC integration does not yet support distributed training with custom weights."
                raise ValueError(msg)

            sampler = provided_sampler or sampler

            shuffle = kwargs.pop("shuffle", True) and sampler is None

            super().__init__(*args, sampler=sampler, shuffle=shuffle, **kwargs)

    # Temporarily patch the InfiniteDataLoader with one that sets the sampler
    ultralytics.data.build.InfiniteDataLoader = _InfiniteDataLoaderWithSampler

    dataloader = build_dataloader_ultralytics(*args, **kwargs)

    # Restore the original InfiniteDataLoader
    ultralytics.data.build.InfiniteDataLoader = InfiniteDataLoader

    return dataloader



from __future__ import annotations

import contextlib
import random

from ultralytics.utils import LOGGER

from tlc_ultralytics.constants import TLC_COLORSTR


def _complete_label_column_name(label_column_name: str, default_label_column_name: str) -> str:
    """Create a complete label column name from a potentially partial one.

    Examples:
        >>> _complete_label_column_name("a", "a")
        "a"
        >>> _complete_label_column_name("a", "a.b.c")
        "a.b.c"
        >>> _complete_label_column_name("a.b.c", "d.e.f")
        "a.b.c"
        >>> _complete_label_column_name("", "a.b.c")
        "a.b.c"
    """
    parts = label_column_name.split(".") if label_column_name else []
    default_parts = default_label_column_name.split(".")

    for i, default_part in enumerate(default_parts):
        if i >= len(parts):
            parts.append(default_part)

    return ".".join(parts)


@contextlib.contextmanager
def _restore_random_state():
    """Context manager to ensure the global random state is unchanged by the wrapped code."""
    state = random.getstate()
    yield
    random.setstate(state)


def _handle_deprecated_column_name(
    arg_value: str | None, settings_value: str | None, default_value: str, column_name: str
) -> str:
    """Handling for when a column name is passed as an argument directly, instead of through a `Settings` object.
    Used in the `trainer` and `validator` classes. A warning is logged if the column name is passed as an argument.

    If the column name is provided both through the `Settings` object and as an argument, the one passed directly is
    used and a warning is logged.

    If the column name is not provided, the default value is returned.

    :param arg_value: The column name passed as an argument directly.
    :param settings_value: The column name set in the `Settings` object.
    :param default_value: The default column name.
    :return: The column name.
    """
    if arg_value is not None:
        msg = (
            f"Passing `{column_name}` as an argument is deprecated. Provide `{column_name}` to a `Settings` object "
            "instead."
        )
        LOGGER.warning(f"{TLC_COLORSTR}{msg}")

        if settings_value is not None:
            msg = (
                f"`{column_name}` is both set in the `Settings` object and provided directly. Using the one passed "
                f"directly: '{arg_value}'."
            )
            LOGGER.warning(f"{TLC_COLORSTR}{msg}")

        return arg_value
    elif settings_value is None:
        return default_value
    return settings_value



from __future__ import annotations

import tlc

from tlc_ultralytics.constants import TRAINING_PHASE


def training_phase_schema() -> tlc.Schema:
    """Create a 3LC schema for the training phase.

    :returns: The training phase schema.
    """
    return tlc.Schema(
        display_name=TRAINING_PHASE,
        description=(
            "'During' metrics are collected with EMA during training, "
            "'After' is with the final model weights after completed training."
        ),
        display_importance=tlc.DISPLAY_IMPORTANCE_EPOCH - 1,  # Right hand side of epoch in the Dashboard
        writable=False,
        computable=False,
        value=tlc.Int32Value(
            value_min=0,
            value_max=1,
            value_map={
                float(0): tlc.MapElement(display_name="During"),
                float(1): tlc.MapElement(display_name="After"),
            },
        ),
    )


def image_embeddings_schema(activation_size=512) -> dict[str, tlc.Schema]:
    """Create a 3LC schema for YOLO image embeddings.

    :param activation_size: The size of the activation tensor.
    :returns: The YOLO image embeddings schema.
    """
    embedding_schema = tlc.Schema(
        "Embedding",
        "Large NN embedding",
        writable=False,
        computable=False,
        value=tlc.Float32Value(number_role=tlc.NUMBER_ROLE_NN_EMBEDDING),
        size0=tlc.DimensionNumericValue(
            value_min=activation_size,
            value_max=activation_size,
            enforce_min=True,
            enforce_max=True,
        ),
    )
    return embedding_schema



from __future__ import annotations

from copy import deepcopy

from ultralytics.models.yolo.pose.train import PoseTrainer
from ultralytics.nn.tasks import PoseModel
from ultralytics.utils import LOGGER
from ultralytics.utils.loss import E2ELoss

from tlc_ultralytics.constants import IMAGE_COLUMN_NAME, POSE_LABEL_COLUMN_NAME, TLC_COLORSTR
from tlc_ultralytics.detect.trainer import TLCDetectionTrainer
from tlc_ultralytics.pose.loss import TLCPoseLoss26, TLCv8PoseLoss
from tlc_ultralytics.pose.validator import TLCPoseValidator
from tlc_ultralytics.utils.dataset import check_tlc_dataset


# Monkeypatch Ultralytics PoseModel to use TLC pose losses which respect dataset oks_sigmas
def _tlc_pose_init_criterion(self):
    return E2ELoss(self, TLCPoseLoss26) if getattr(self, "end2end", False) else TLCv8PoseLoss(self)


# Apply the monkeypatch once at import time
PoseModel.init_criterion = _tlc_pose_init_criterion


class TLCPoseTrainer(PoseTrainer, TLCDetectionTrainer):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = POSE_LABEL_COLUMN_NAME

    def get_dataset(self):
        self.data = check_tlc_dataset(
            self.args.data,
            self._tables,
            self._settings.image_column_name,
            self._settings.label_column_name,
            project_name=self._settings.project_name,
            splits=("train", "val"),
            task="pose",
            settings=self._settings,
        )

        # Get test data if val not present
        if "val" not in self.data:
            data_test = check_tlc_dataset(
                self.args.data,
                self._tables,
                self._settings.image_column_name,
                self._settings.label_column_name,
                project_name=self._settings.project_name,
                splits=("test",),
                task="pose",
                settings=self._settings,
            )
            self.data["test"] = data_test["test"]

        return self.data

    def set_model_attributes(self):
        """Set keypoints shape and attach dataset-provided OKS sigmas to the model if available."""
        super().set_model_attributes()
        oks_sigmas = self._settings.oks_sigmas or self.data.get("oks_sigmas")
        if oks_sigmas is not None:
            # Attach to model so TLCv8PoseLoss/v8UnreducedPoseLoss can pick it up
            self.model.oks_sigmas = oks_sigmas

    def get_validator(self, dataloader=None):
        self.loss_names = ("box_loss", "pose_loss", "kobj_loss", "cls_loss", "dfl_loss")
        if not dataloader:
            dataloader = self.test_loader

        return TLCPoseValidator(
            dataloader,
            save_dir=self.save_dir,
            args=deepcopy(self.args),
            _callbacks=self.callbacks,
            run=self._run,
            settings=self._settings,
            training=True,
        )

    def _process_metrics(self, metrics):
        detection_metrics = super()._process_metrics(metrics)
        return {metric_name.replace("(P)", "_pose"): value for metric_name, value in detection_metrics.items()}

    def _print_task_specific_parameters(self):
        """Print task-specific parameters to the console."""
        table_sigmas = self.data.get("oks_sigmas")

        if table_sigmas is not None:
            table_sigmas_rounded = [round(x, 2) for x in table_sigmas]
            LOGGER.info(f"{TLC_COLORSTR}Using OKS sigmas: {table_sigmas_rounded} from Table for evaluation")
        if self._settings.oks_sigmas is not None:
            settings_oks_rounded = [round(x, 2) for x in self._settings.oks_sigmas]
            LOGGER.info(
                f"{TLC_COLORSTR}Using overridden OKS sigmas: {settings_oks_rounded} from Settings for computing loss"
            )



from __future__ import annotations

import tlc
from tlc.client.torch.metrics.metrics_collectors.bounding_box_metrics_collector import (
    _TLCPredictedBoundingBox,
    _TLCPredictedBoundingBoxes,
)

from tlc_ultralytics.detect.dataset import TLCYOLODataset
from tlc_ultralytics.settings import Settings


def get_or_create_det_table(
    key: str,
    data_dict: dict[str, object],
    image_column_name: str,
    label_column_name: str,
    project_name: str,
    dataset_name: str,
    table_name: str,
    settings: Settings | None = None,
) -> tlc.Table:
    """Get or create a detection table from a dataset dictionary.

    :param key: The key of the dataset dictionary (the split to use)
    :param data_dict: Dictionary of dataset information
    :param project_name: Name of the project
    :param dataset_name: Name of the dataset
    :param table_name: Name of the table
    :param image_column_name: Name of the column containing image paths
    :param label_column_name: Name of the column containing labels
    :return: A tlc.Table.from_yolo() table
    """
    return tlc.Table.from_yolo(
        dataset_yaml_file=data_dict["yaml_file"],
        split=key,
        override_split_path=data_dict[key],
        task="detect",
        project_name=project_name,
        dataset_name=dataset_name,
        table_name=table_name,
        if_exists="reuse",
        add_weight_column=True,
        description="Created with 3LC YOLO integration",
    )


def build_tlc_yolo_dataset(
    cfg,
    table,
    batch,
    data,
    mode="train",
    rect=False,
    stride=32,
    multi_modal=False,
    exclude_zero=False,
    class_map=None,
    split=None,
    image_column_name=None,
    label_column_name=None,
):
    if multi_modal:
        return ValueError("Multi-modal datasets are not supported in the 3LC Ultralytics integration.")

    return TLCYOLODataset(
        table,
        exclude_zero=exclude_zero,
        class_map=class_map,
        imgsz=cfg.imgsz,
        batch_size=batch,
        augment=mode == "train",  # augmentation
        hyp=cfg,  # TODO: probably add a get_hyps_from_cfg function
        rect=cfg.rect or rect,  # rectangular batches
        cache=cfg.cache or None,
        single_cls=cfg.single_cls or False,
        stride=int(stride),
        pad=0.0 if mode == "train" else 0.5,
        prefix=split or mode,
        task=cfg.task,
        classes=cfg.classes,
        data=data,
        fraction=cfg.fraction if mode == "train" else 1.0,
        image_column_name=image_column_name,
        label_column_name=label_column_name,
    )


def check_det_table(
    table: tlc.Table,
    image_column_name: str = tlc.IMAGE,
    label_column_name: str = f"{tlc.BOUNDING_BOXES}.{tlc.BOUNDING_BOX_LIST}.{tlc.LABEL}",
) -> None:
    """Check that a table is compatible with the detection task in the 3LC YOLO integration.

    :param table: The table to check.
    :param image_column_name: The name of the column containing image paths.
    :param label_column_name: The full label path of the column containing labels.
    :raises: ValueError if the table is not compatible with the detection task.
    """
    row_schema = table.row_schema.values

    bounding_boxes_column_key, bounding_boxes_list_key, label_key = label_column_name.split(".")

    try:
        assert image_column_name in row_schema, f"Image column '{image_column_name}' not found."
        assert bounding_boxes_column_key in row_schema, f"Bounding box column '{bounding_boxes_column_key}' not found."
        assert bounding_boxes_list_key in row_schema[bounding_boxes_column_key].values, (
            f"Bounding box list '{bounding_boxes_list_key}' not found in column '{bounding_boxes_column_key}'."
        )

        assert tlc.IMAGE_HEIGHT in row_schema[bounding_boxes_column_key].values, (
            f"Bounding box column '{bounding_boxes_column_key}' does not contain a key '{tlc.IMAGE_HEIGHT}'."
        )
        assert tlc.IMAGE_WIDTH in row_schema[bounding_boxes_column_key].values, (
            f"Bounding box column '{bounding_boxes_column_key}' does not contain a key '{tlc.IMAGE_WIDTH}'."
        )

        for coordinate in [tlc.X0, tlc.Y0, tlc.X1, tlc.Y1]:
            assert coordinate in row_schema[bounding_boxes_column_key].values[bounding_boxes_list_key].values, (
                f"Bounding box list '{bounding_boxes_list_key}' in column '{bounding_boxes_column_key}' "
                f"does not contain a key '{coordinate}'."
            )
        assert label_key in row_schema[bounding_boxes_column_key].values[bounding_boxes_list_key].values, (
            f"Bounding box list '{bounding_boxes_list_key}' in column '{bounding_boxes_column_key}' "
            f"does not contain a key '{label_key}'."
        )
        assert table.get_value_map(label_column_name) is not None, (
            f"Unable to get value map for label value path {label_column_name}. Ensure that the table is compatible "
            "with the detection task or provide a `label_column_name` that matches the value path to the labels."
        )

    except (AssertionError, KeyError) as e:
        raise ValueError(f"Table with url {table.url} is not compatible with YOLO object detection. {e}") from None


def yolo_predicted_bounding_box_schema(
    label_value_map: dict[float, tlc.MapElement],
) -> tlc.Schema:
    """Create a 3LC bounding box schema for YOLO.

    :param categories: Categories for the current dataset.
    :returns: The YOLO bounding box schema for predicted boxes.
    """

    bounding_box_schema = tlc.BoundingBoxListSchema(
        label_value_map,
        x0_number_role=tlc.NUMBER_ROLE_BB_CENTER_X,
        x1_number_role=tlc.NUMBER_ROLE_BB_SIZE_X,
        y0_number_role=tlc.NUMBER_ROLE_BB_CENTER_Y,
        y1_number_role=tlc.NUMBER_ROLE_BB_SIZE_Y,
        x0_unit=tlc.UNIT_RELATIVE,
        y0_unit=tlc.UNIT_RELATIVE,
        x1_unit=tlc.UNIT_RELATIVE,
        y1_unit=tlc.UNIT_RELATIVE,
        description="Predicted Bounding Boxes",
        writable=False,
        is_prediction=True,
        include_segmentation=False,
    )

    return bounding_box_schema


def yolo_loss_schemas(training: bool = False) -> dict[str, tlc.Schema]:
    """Create a 3LC schema for YOLO per-sample loss metrics.

    :param training: Whether metrics are collected during training.
    :returns: The YOLO loss schemas for each of the three components.
    """
    schemas = {}
    schemas["box_loss"] = tlc.Schema(
        description="Box Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3004,
    )
    schemas["dfl_loss"] = tlc.Schema(
        description="Distribution Focal Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3005,
    )
    schemas["cls_loss"] = tlc.Schema(
        description="Classification Loss",
        writable=False,
        value=tlc.Float32Value(),
        display_importance=3006,
    )
    if training:
        schemas["loss"] = tlc.Schema(
            description="Weighted sum of box, DFL, and classification losses used in training",
            writable=False,
            value=tlc.Float32Value(),
            display_importance=3007,
        )
    return schemas


def construct_bbox_struct(
    predicted_annotations: list[dict[str, int | float | dict[str, float]]],
    image_width: int,
    image_height: int,
    inverse_label_mapping: dict[int, int] | None = None,
) -> _TLCPredictedBoundingBoxes:
    """Construct a 3LC bounding box struct from a list of bounding boxes.

    :param predicted_annotations: A list of predicted bounding boxes.
    :param image_width: The width of the image.
    :param image_height: The height of the image.
    :param inverse_label_mapping: A mapping from predicted label to category id.
    """

    bbox_struct = _TLCPredictedBoundingBoxes(
        bb_list=[],
        image_width=image_width,
        image_height=image_height,
    )

    for pred in predicted_annotations:
        bbox, label, score, iou = (
            pred["bbox"],
            pred["category_id"],
            pred["score"],
            pred["iou"],
        )
        label_val = inverse_label_mapping[label] if inverse_label_mapping is not None else label
        bbox_struct["bb_list"].append(
            _TLCPredictedBoundingBox(
                label=label_val,
                confidence=score,
                iou=iou,
                x0=bbox[0],
                y0=bbox[1],
                x1=bbox[2],
                y1=bbox[3],
            )
        )

    return bbox_struct
